import MuyGPyS._src.math.torch as torch


def _rbf_fn(squared_dists: torch.ndarray, **kwargs) -> torch.ndarray:
    return torch.exp(-squared_dists / 2.0)


def _matern_05_fn(dists: torch.ndarray, **kwargs) -> torch.ndarray:
    return torch.exp(-dists)


def _matern_15_fn(dists: torch.ndarray, **kwargs) -> torch.ndarray:
    K = dists * torch.sqrt(torch.array(3))
    return (1.0 + K) * torch.exp(-K)


def _matern_25_fn(dists: torch.ndarray, **kwargs) -> torch.ndarray:
    K = dists * torch.sqrt(torch.array(5))
    return (1.0 + K + K**2 / 3.0) * torch.exp(-K)


def _matern_inf_fn(dists: torch.ndarray, **kwargs) -> torch.ndarray:
    return torch.exp(-(dists**2) / 2.0)


def _matern_gen_fn(dists: torch.ndarray, nu: float, **kwargs) -> torch.ndarray:
    raise NotImplementedError(
        f'Function "matern_gen_fn" does not support nu={nu}. '
        f"Torch only supports 1/2, 3/2, 5/2 and torch.inf."
    )



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from MuyGPyS._src.gp.noise.numpy import (
    _homoscedastic_perturb,
    _heteroscedastic_perturb,
)



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from MuyGPyS import config


def _collect_implementation(package, *funcs):
    if config.state.backend == "numpy":
        return _collect_functions(package + ".numpy", *funcs)
    elif config.state.backend == "jax":
        return _collect_functions(package + ".jax", *funcs)
    elif config.state.backend == "torch":
        return _collect_functions(package + ".torch", *funcs)
    elif config.state.backend == "mpi":
        return _collect_functions(package + ".mpi", *funcs)
    else:
        raise ValueError(
            f'MuyGPyS backend is in bad state "{config.state.backend}"'
        )


def _fullname(klass):
    module = klass.__module__
    if module == "builtins":
        return klass.__qualname__
    return module + "." + klass.__qualname__


def _collect_functions(package, *funcs):
    return tuple([getattr(__import__(package, fromlist=[f]), f) for f in funcs])


def auto_str(klass):
    def __str__(self):
        public_members = ", ".join(
            "%s=%s" % item
            for item in vars(self).items()
            if not item[0].startswith("_")
        )
        return f"{type(self).__name__}({public_members})"

    klass.__str__ = __str__
    return klass



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

import MuyGPyS._src.math.numpy as np


def _analytic_sigma_sq_optim_unnormalized(
    K: np.ndarray,
    nn_targets: np.ndarray,
) -> np.ndarray:
    return np.sum(
        np.einsum("ijk,ijk->ik", nn_targets, np.linalg.solve(K, nn_targets)),
        axis=0,
    )


def _analytic_sigma_sq_optim(
    K: np.ndarray,
    nn_targets: np.ndarray,
) -> np.ndarray:
    batch_count, nn_count, _ = nn_targets.shape
    return _analytic_sigma_sq_optim_unnormalized(K, nn_targets) / (
        nn_count * batch_count
    )



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from .kernel_fn import KernelFn
from .matern import Matern
from .rbf import RBF


def _get_kernel(kern: str, **kwargs) -> KernelFn:
    """
    Select and return an appropriate kernel functor based upon the passed
    parameters.

    Args:
        kern:
            The kernel function to be used. Current supports only `"matern"` and
            `"rbf"`.
        kwargs : dict
            Kernel parameters, possibly including hyperparameter dicts.

    Return:
        The appropriately initialized kernel functor.
    """
    if kern == "rbf":
        return RBF(**kwargs)
    elif kern == "matern":
        return Matern(**kwargs)
    else:
        raise ValueError(f"Kernel type {kern} is not supported!")



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

"""
MuyGPs implementation
"""

from typing import Callable, Union

import MuyGPyS._src.math as mm
from MuyGPyS._src.gp.muygps import _muygps_diagonal_variance
from MuyGPyS.gp.sigma_sq import SigmaSq, sigma_sq_scale, sigma_sq_apply
from MuyGPyS.gp.noise import (
    HomoscedasticNoise,
    HeteroscedasticNoise,
    NullNoise,
    perturb_with_noise_model,
)


class PosteriorVariance:
    def __init__(
        self,
        eps: Union[HomoscedasticNoise, HeteroscedasticNoise, NullNoise],
        sigma_sq: SigmaSq,
        apply_sigma_sq=True,
        **kwargs,
    ):
        self.eps = eps
        self.sigma_sq = sigma_sq
        self._fn = _muygps_diagonal_variance
        self._fn = perturb_with_noise_model(self._fn, self.eps)
        if apply_sigma_sq is True:
            self._fn = sigma_sq_scale(self._fn)

    def __call__(
        self,
        K: mm.ndarray,
        Kcross: mm.ndarray,
    ) -> mm.ndarray:
        return self._fn(K, Kcross, eps=self.eps(), sigma_sq=self.sigma_sq())

    def get_opt_fn(self) -> Callable:
        return self._get_opt_fn(self._fn, self.eps, self.sigma_sq)

    @staticmethod
    def _get_opt_fn(
        var_fn: Callable,
        eps: Union[HomoscedasticNoise, HeteroscedasticNoise, NullNoise],
        sigma_sq: SigmaSq,
    ) -> Callable:
        opt_fn = eps.apply(var_fn, "eps")
        opt_fn = sigma_sq_apply(opt_fn, sigma_sq)
        return opt_fn



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT


from scipy.special import gamma, kv

import MuyGPyS._src.math.numpy as np


def _rbf_fn(squared_dists: np.ndarray, **kwargs) -> np.ndarray:
    return np.exp(-squared_dists / 2.0)


def _matern_05_fn(dists: np.ndarray, **kwargs) -> np.ndarray:
    return np.exp(-dists)


def _matern_15_fn(dists: np.ndarray, **kwargs) -> np.ndarray:
    K = dists * np.sqrt(3)
    return (1.0 + K) * np.exp(-K)


def _matern_25_fn(dists: np.ndarray, **kwargs) -> np.ndarray:
    K = dists * np.sqrt(5)
    return (1.0 + K + K**2 / 3.0) * np.exp(-K)


def _matern_inf_fn(dists: np.ndarray, **kwargs) -> np.ndarray:
    return np.exp(-(dists**2) / 2.0)


def _matern_gen_fn(dists: np.ndarray, nu: float, **kwargs) -> np.ndarray:
    K = dists
    K[K == 0.0] += np.finfo(float).eps
    tmp = np.sqrt(2 * nu) * K
    K.fill((2 ** (1.0 - nu)) / gamma(nu))
    K *= tmp**nu
    K *= kv(nu, tmp)
    return K



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

"""
Shear kernel functor

Defines the lensing shear kernel (inheriting
:class:`~MuyGPyS.gp.kernels.kernel_fn.KernelFn`) that transform crosswise and
pairwise difference tensors into cross-covariance and covariance (or kernel)
tensors, respectively.

This kernel is defined in terms of second-order partial derivatives of the RBF
kernel. Mathematical derivation is forthcoming.

Example:
    >>> shear_fn = ShearKenrel(
    ...     metric=IsotropicDistortion(
    ...         F2,
    ...         length_scale=ScalarHyperparameter(1.0),
    ...     ),
    ... )

One uses a previously computed `pairwise_diffs` tensor (see
:func:`MuyGPyS.gp.tensors.pairwise_tensor`) to compute a kernel tensor whose
second two dimensions contain square kernel matrices. Similarly, one uses a
previously computed `crosswise_diffs` matrix (see
:func:`MuyGPyS.gp.tensors.crosswise_tensor`) to compute a cross-covariance
matrix. See the following example, which assumes that you have already
constructed the difference `numpy.nparrays` and the kernel `kernel_fn` as shown
above.

Example:
    >>> K = shear_fn(pairwise_diffs)
    >>> Kcross = shear_fn(crosswise_diffs)

Note that the :class:`~MuyGPyS.gp.kernels.experimental.shear.ShearKernel`
functor
"""

from typing import Callable, List, Tuple

import MuyGPyS._src.math as mm
from MuyGPyS._src.gp.kernels.shear import _shear_fn
from MuyGPyS._src.util import auto_str
from MuyGPyS.gp.distortion import (
    IsotropicDistortion,
    F2,
)
from MuyGPyS.gp.kernels import KernelFn
from MuyGPyS.gp.hyperparameter import ScalarHyperparameter


@auto_str
class ShearKernel(KernelFn):
    """
    The lensing shear kernel.

    The lensing shear kernel is defined in terms of second-order partial
    derivates of the RBF kernel.

    The kernel is defined by

    .. math::
        K(x_i, x_j) = \\dots

    Args:
        metric:
            The distance function to be used. Includes length_scale
            hyperparameter information via the MuyGPyS.gp.distortion module
    """

    def __init__(
        self,
        metric: IsotropicDistortion = IsotropicDistortion(
            F2, length_scale=ScalarHyperparameter(1.0)
        ),
    ):
        super().__init__(metric=metric)
        if not isinstance(self.distortion_fn, IsotropicDistortion):
            raise ValueError(
                f"ShearKernel only supports isotropic metrics, not {type(metric)}"
            )
        self._fn = _shear_fn
        self._make()

    def _make(self):
        super()._make_base()

    def __call__(self, diffs: mm.ndarray, **kwargs) -> mm.ndarray:
        """
        Compute shear kernel(s) from a difference tensor.

        Args:
            diffs:
                A tensor of pairwise diffs of shape
                `(data_count, nn_count, nn_count, feature_count)` or
                `(data_count, nn_count, feature_count)`. In the four dimensional
                case, it is assumed that the diagonals dists
                diffs[i, j, j, :] == 0.
            batch_features:
                A tensor of shape `(data_count, 1)` or a vector of length
                `data_count` or a scalar.

        Returns:
            A cross-covariance matrix of shape `(data_count * 3, nn_count * 3)`
            or a tensor of shape `(data_count, nn_count * 3, nn_count * 3)`
            whose last two dimensions are kernel matrices.
        """
        return self._fn(
            diffs, length_scale=self.distortion_fn.length_scale(), **kwargs
        )

    def get_opt_params(
        self,
    ) -> Tuple[List[str], List[float], List[Tuple[float, float]]]:
        """
        Report lists of unfixed hyperparameter names, values, and bounds.

        Returns
        -------
            names:
                A list of unfixed hyperparameter names.
            params:
                A list of unfixed hyperparameter values.
            bounds:
                A list of unfixed hyperparameter bound tuples.
        """
        names, params, bounds = super().get_opt_params()
        return names, params, bounds

    def get_opt_fn(self) -> Callable:
        """
        Return a kernel function with fixed parameters set.

        This function is designed for use with
        :func:`MuyGPyS.optimize.chassis.optimize_from_tensors()` and assumes
        that optimization parameters will be passed as keyword arguments.

        Returns:
            A function implementing the kernel where all fixed parameters are
            set. The function expects keyword arguments corresponding to current
            hyperparameter values for unfixed parameters.
        """
        return super()._get_opt_fn(self._fn, self.distortion_fn)



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from copy import deepcopy
from typing import Callable, List, Tuple

from bayes_opt import BayesianOptimization
from scipy import optimize as opt

from MuyGPyS.gp import MuyGPS


def _new_muygps(muygps: MuyGPS, x0_names, bounds, opt_dict) -> MuyGPS:
    # This will need to be revisited once we start using heteroscedastic noise
    ret = deepcopy(muygps)
    for i, key in enumerate(x0_names):
        lb, ub = bounds[i]
        val = opt_dict[key]
        if val < lb:
            val = lb
        elif val > ub:
            val = ub
        if key == "eps":
            ret.eps._set_val(val)
        elif "_knot" in key:
            loc = key.find("_knot")
            name = key[:loc]
            ret.kernel._hyperparameters[name]._update_knot_values(**{key: val})
        else:
            ret.kernel._hyperparameters[key]._set_val(val)
    ret._make()
    return ret


def _obj_fn_adapter(obj_fn, x0_names):
    def array_obj_fn(x_array, *args):
        arr_dict = {h: x_array[i] for i, h in enumerate(x0_names)}
        return -obj_fn(*args, **arr_dict)

    return array_obj_fn


def _get_opt_lists(
    muygps: MuyGPS,
    verbose: bool = False,
) -> Tuple[List[str], List[float], List[Tuple[float, float]]]:
    x0_names, x0, bounds = muygps.get_opt_params()
    if verbose is True:
        print(f"parameters to be optimized: {x0_names}")
        print(f"bounds: {bounds}")
        print(f"initial x0: {x0}")
    return x0_names, x0, bounds


def _scipy_optimize(
    muygps: MuyGPS,
    obj_fn: Callable,
    verbose: bool = False,
    **kwargs,
) -> MuyGPS:
    x0_names, x0, bounds = _get_opt_lists(muygps, verbose=verbose)

    # converting from kwargs representation to array representation
    array_obj_fn = _obj_fn_adapter(obj_fn, x0_names)

    optres = opt.minimize(
        array_obj_fn,
        x0,
        method="L-BFGS-B",
        bounds=bounds,
        **kwargs,
    )
    if verbose is True:
        print(f"optimizer results: \n{optres}")

    # converting back to kwargs representation
    ret_dict = {n: optres.x[i] for i, n in enumerate(x0_names)}

    return _new_muygps(muygps, x0_names, bounds, ret_dict)


def _bayes_get_kwargs(verbose: bool = False, **kwargs):
    optimizer_kwargs = {
        k: kwargs[k]
        for k in kwargs
        if k
        in {
            "random_state",
            "verbose",
            "bounds_transformer",
            "allow_duplicate_points",
        }
    }
    if "verbose" not in optimizer_kwargs:
        optimizer_kwargs["verbose"] = 2 if verbose is True else 0

    # not allowing the user to set the opt gp parameters for now
    maximize_kwargs = {
        k: kwargs[k]
        for k in kwargs
        if k
        in {
            "init_points",
            "n_iter",
            "acq",
            "kappa",
            "kappa_decay",
            "kappa_decay_delay",
            "xi",
        }
    }
    return optimizer_kwargs, maximize_kwargs


def _bayes_opt_optimize(
    muygps: MuyGPS,
    obj_fn: Callable,
    verbose: bool = False,
    **kwargs,
) -> MuyGPS:
    x0_names, x0, bounds = _get_opt_lists(muygps, verbose=verbose)

    x0_map = {n: x0[i] for i, n in enumerate(x0_names)}
    bounds_map = {n: bounds[i] for i, n in enumerate(x0_names)}

    optimizer_kwargs, maximize_kwargs = _bayes_get_kwargs(
        verbose=verbose,
        **kwargs,
    )

    # set defaults
    if "init_points" not in maximize_kwargs:
        maximize_kwargs["init_points"] = 5
    if "n_iter" not in maximize_kwargs:
        maximize_kwargs["n_iter"] = 20

    optimizer = BayesianOptimization(
        f=obj_fn,
        pbounds=bounds_map,
        **optimizer_kwargs,
    )
    optimizer.probe(x0_map, lazy=True)
    optimizer.maximize(**maximize_kwargs)

    return _new_muygps(muygps, x0_names, bounds, optimizer.max["params"])



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from typing import Dict, Callable, Union

from MuyGPyS.gp.hyperparameter import ScalarHyperparameter
from MuyGPyS.gp.hyperparameter.experimental import (
    HierarchicalNonstationaryHyperparameter,
)


class NullDistortion:
    def __init__(
        self,
        metric: Callable,
        length_scale: Union[
            ScalarHyperparameter, HierarchicalNonstationaryHyperparameter
        ],
    ):
        self.length_scale = length_scale
        self._dist_fn = metric

    def __call__(self, *args, **kwargs):
        raise NotImplementedError("NullDistortion cannot be called!")

    def populate_length_scale(
        self, hyperparameters: Dict, *args, **kwargs
    ) -> None:
        """
        A no-op placeholder function for NullDistortion objects.
        """

    def get_opt_fn(self, *args, **kwargs):
        raise NotImplementedError(
            "NullDistortion cannot be used in optimization!"
        )

    def get_opt_params(self, *args, **kwargs):
        raise NotImplementedError(
            "NullDistortion cannot be used in optimization!"
        )



# Copyright 2023-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from .hierarchical_nonstationary import (
    HierarchicalNonstationaryHyperparameter,
    sample_knots,
)



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

import MuyGPyS._src.math.torch as torch


def _cross_entropy_fn(
    predictions: torch.ndarray,
    targets: torch.ndarray,
) -> torch.ndarray:
    one_hot_targets = torch.where(targets > 0.0, 1.0, 0.0)
    softmax_predictions = predictions.softmax(dim=1)

    return (
        -torch.mean(
            one_hot_targets * torch.log(softmax_predictions)
            + (1 - one_hot_targets) * torch.log(1 - softmax_predictions)
        )
        * one_hot_targets.shape[0]
    )


def _mse_fn_unnormalized(
    predictions: torch.ndarray,
    targets: torch.ndarray,
) -> float:
    return torch.sum((predictions - targets) ** 2)


def _mse_fn(
    predictions: torch.ndarray,
    targets: torch.ndarray,
) -> float:
    batch_count, response_count = predictions.shape
    return _mse_fn_unnormalized(predictions, targets) / (
        batch_count * response_count
    )


def _lool_fn(
    predictions: torch.ndarray,
    targets: torch.ndarray,
    variances: torch.ndarray,
    sigma_sq: torch.ndarray,
) -> float:
    return _lool_fn_unscaled(
        predictions, targets, torch.unsqueeze(variances * sigma_sq, dim=1)
    )


def _lool_fn_unscaled(
    predictions: torch.ndarray, targets: torch.ndarray, variances: torch.ndarray
) -> float:
    return torch.sum(
        torch.divide((predictions - targets) ** 2, variances)
        + torch.log(variances)
    )


def _pseudo_huber_fn(
    predictions: torch.ndarray,
    targets: torch.ndarray,
    boundary_scale: float = 1.5,
) -> float:
    return boundary_scale**2 * torch.sum(
        torch.sqrt(1 + torch.divide(targets - predictions, boundary_scale) ** 2)
        - 1
    )


def _looph_fn_unscaled(
    predictions: torch.ndarray,
    targets: torch.ndarray,
    variances: torch.ndarray,
    boundary_scale: float = 1.5,
) -> float:
    boundary_scale_sq = boundary_scale**2
    return torch.sum(
        boundary_scale_sq
        * torch.sqrt(
            1
            + torch.divide(targets - predictions, variances * boundary_scale)
            ** 2
        )
        - boundary_scale_sq
        + torch.log(variances)
    )


def _looph_fn(
    predictions: torch.ndarray,
    targets: torch.ndarray,
    variances: torch.ndarray,
    sigma_sq: torch.ndarray,
    boundary_scale: float = 1.5,
) -> float:
    return _looph_fn_unscaled(
        predictions,
        targets,
        torch.outer(variances, sigma_sq),
        boundary_scale=boundary_scale,
    )



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT


from typing import Callable, Dict, List, Tuple, Union

import MuyGPyS._src.math as mm
from MuyGPyS._src.util import auto_str
from MuyGPyS.gp.hyperparameter import ScalarHyperparameter
from MuyGPyS.gp.hyperparameter.experimental import (
    HierarchicalNonstationaryHyperparameter,
)


@auto_str
class IsotropicDistortion:
    """
    An isotropic distance model.

    IsotropicDistortion parameterizes a scaled elementwise distance function
    :math:`d(\\cdot, \\cdot)`, and is paramterized by a scalar :math:`\\ell>0`
    length scale hyperparameter.

    .. math::
         d_\\ell(\\mathbf{x}, \\mathbf{y}) =
         \\sum_{i=0}^d \\frac{d(\\mathbf{x}_i, \\mathbf{y}_i)}{\\ell}

    Args:
        metric:
            A callable metric function that takes a tensor of shape
            `(..., feature_count)` whose last dimension lists the elementwise
            differences between a pair of feature vectors and returns a tensor
            of shape `(...)`, having collapsed the last dimension into a
            scalar difference.
        length_scale:
            Some scalar nonnegative hyperparameter object.
    """

    def __init__(
        self,
        metric: Callable,
        length_scale: Union[
            ScalarHyperparameter, HierarchicalNonstationaryHyperparameter
        ],
    ):
        self.length_scale = length_scale
        self._dist_fn = metric

    def __call__(
        self, diffs: mm.ndarray, length_scale: Union[float, mm.ndarray]
    ) -> mm.ndarray:
        """
        Apply isotropic distortion to an elementwise difference tensor.

        This function is not intended to be invoked directly by a user. It is
        instead functionally incorporated into some
        :class:`MuyGPyS.gp.kernels.KernelFn` in its constructor.

        Args:
            diffs:
                A tensor of pairwise differences of shape
                `(..., feature_count)`.
            length_scale:
                A floating point length scale, or a vector of `(knot_count,)`
                knot length scales.
        Returns:
            A crosswise distance matrix of shape `(data_count, nn_count)` or a
            pairwise distance tensor of shape
            `(data_count, nn_count, nn_count)` whose last two dimensions are
            pairwise distance matrices.
        """
        length_scale_array = self._get_length_scale_array(
            diffs.shape, length_scale
        )
        return self._dist_fn(diffs / length_scale_array)

    @staticmethod
    def _get_length_scale_array(
        target_shape: mm.ndarray,
        length_scale: Union[float, mm.ndarray],
    ) -> mm.ndarray:
        # make sure length_scale is broadcastable when its shape is (batch_count,)
        # NOTE[MWP] there is probably a better way to do this
        shape = (-1,) + (1,) * (len(target_shape) - 1)
        return mm.reshape(mm.promote(length_scale), shape)

    def get_opt_params(
        self,
    ) -> Tuple[List[str], List[float], List[Tuple[float, float]]]:
        """
        Report lists of unfixed hyperparameter names, values, and bounds.

        Returns
        -------
            names:
                A list of unfixed hyperparameter names.
            params:
                A list of unfixed hyperparameter values.
            bounds:
                A list of unfixed hyperparameter bound tuples.
        """
        names: List[str] = []
        params: List[float] = []
        bounds: List[Tuple[float, float]] = []
        self.length_scale.append_lists("length_scale", names, params, bounds)
        return names, params, bounds

    def get_opt_fn(self, fn) -> Callable:
        """
        Return a kernel function with fixed parameters set.

        This function is designed for use with
        :func:`MuyGPyS.optimize.chassis.optimize_from_tensors()` and assumes
        that optimization parameters will be passed as keyword arguments.

        Returns:
            A function implementing the kernel where all fixed parameters are
            set. The function expects keyword arguments corresponding to current
            hyperparameter values for unfixed parameters.
        """
        opt_fn = self.length_scale.apply(fn, "length_scale")
        return opt_fn

    def populate_length_scale(self, hyperparameters: Dict) -> None:
        """
        Populates the hyperparameter dictionary of a KernelFn object with
        `self.length_scale` of the IsotropicDistortion object.

        Args:
            hyperparameters:
                A dict containing the hyperparameters of a KernelFn object.
        """
        hyperparameters["length_scale"] = self.length_scale



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from typing import Callable, Dict, List, Optional, Tuple, Union

from absl.testing import parameterized

import MuyGPyS._src.math.numpy as np
from MuyGPyS._src.mpi_utils import (
    _consistent_chunk_tensor,
    _consistent_unchunk_tensor,
    _consistent_reduce_scalar,
)
from MuyGPyS._test.utils import _check_ndarray
from MuyGPyS.examples.classify import do_classify
from MuyGPyS.examples.two_class_classify_uq import do_classify_uq, do_uq
from MuyGPyS.examples.regress import do_regress
from MuyGPyS.examples.fast_posterior_mean import do_fast_posterior_mean
from MuyGPyS.gp import MuyGPS, MultivariateMuyGPS as MMuyGPS
from MuyGPyS.gp.kernels import Matern
from MuyGPyS.optimize.loss import mse_fn


class APITestCase(parameterized.TestCase):
    def _check_ndarray(self, *args, **kwargs):
        return _check_ndarray(self.assertEqual, args, kwargs)


class ClassifyAPITest(APITestCase):
    def _do_classify_test_chassis(
        self,
        train: Dict[str, np.ndarray],
        test: Dict[str, np.ndarray],
        target_acc: float,
        nn_count: int,
        batch_count: int,
        loss_fn: Callable,
        obj_method: str,
        opt_method: str,
        nn_kwargs: Dict,
        k_kwargs: Union[Dict, Union[List[Dict], Tuple[Dict, ...]]],
        opt_kwargs: Dict,
        verbose: bool = False,
    ) -> None:
        (
            muygps,
            surrogate_predictions,
            predicted_labels,
            acc,
        ) = self._do_classify(
            train,
            test,
            nn_count,
            batch_count,
            loss_fn,
            obj_method,
            opt_method,
            nn_kwargs,
            k_kwargs,
            opt_kwargs,
            verbose=verbose,
        )
        self.assertEqual(surrogate_predictions.shape, test["output"].shape)
        self.assertEqual(predicted_labels.shape, (test["output"].shape[0],))
        if isinstance(k_kwargs, dict):
            self.assertSequenceAlmostEqual(
                np.sum(surrogate_predictions, axis=1),
                np.zeros(surrogate_predictions.shape[0]),
            )
        # There is almost certainly a better way to do this.
        if np.all(np.unique(train["output"]) == np.unique([-1, 1])):
            self.assertSequenceAlmostEqual(
                np.unique(predicted_labels),
                np.unique(train["output"]),
            )
        else:
            self.assertSequenceAlmostEqual(
                np.unique(predicted_labels),
                np.unique(np.argmax(train["output"], axis=1)),
            )
        print("Finds hyperparameters:")
        if isinstance(muygps, MuyGPS):
            param_names, param_vals, _ = muygps.get_opt_params()
            for i, p in enumerate(param_names):
                print(f"\t{p} : {param_vals[i]}")
        elif isinstance(muygps, MMuyGPS):
            for i, model in enumerate(muygps.models):
                print(f"model {i}:")
                param_names, param_vals, _ = model.get_opt_params()
                for i, p in enumerate(param_names):
                    print(f"\t{p} : {param_vals[i]}")
        print(f"obtains accuracy: {acc}")
        self.assertGreaterEqual(acc, target_acc)

    def _do_classify(
        self,
        train: Dict[str, np.ndarray],
        test: Dict[str, np.ndarray],
        nn_count: int,
        batch_count: int,
        loss_fn: Callable,
        obj_method: str,
        opt_method: str,
        nn_kwargs: Dict,
        k_kwargs: Union[Dict, Union[List[Dict], Tuple[Dict, ...]]],
        opt_kwargs: Dict,
        verbose: bool = False,
    ) -> Tuple[Union[MuyGPS, MMuyGPS], np.ndarray, np.ndarray, float]:
        classifier, _, surrogate_predictions = do_classify(
            test["input"],
            train["input"],
            train["output"],
            nn_count=nn_count,
            batch_count=batch_count,
            loss_fn=loss_fn,
            obj_method=obj_method,
            opt_method=opt_method,
            k_kwargs=k_kwargs,
            nn_kwargs=nn_kwargs,
            opt_kwargs=opt_kwargs,
            verbose=verbose,
        )

        predicted_labels = np.argmax(surrogate_predictions, axis=1)
        target_labels = np.argmax(test["output"], axis=1)
        test_count = len(target_labels)
        if np.all(np.unique(train["output"]) == np.unique([-1.0, 1.0])):
            predicted_labels = 2 * predicted_labels - 1
            target_labels = 2 * target_labels - 1

        target_labels = _consistent_chunk_tensor(target_labels)
        correct_count = np.count_nonzero(predicted_labels == target_labels)

        correct_count = _consistent_reduce_scalar(correct_count)
        acc = correct_count / test_count

        surrogate_predictions = _consistent_unchunk_tensor(
            surrogate_predictions
        )
        predicted_labels = _consistent_unchunk_tensor(predicted_labels)
        return (
            classifier,
            surrogate_predictions,
            predicted_labels,
            acc,
        )

    def _do_classify_uq_test_chassis(
        self,
        train: Dict[str, np.ndarray],
        test: Dict[str, np.ndarray],
        target_acc: float,
        nn_count: int,
        opt_batch_count: int,
        uq_batch_count: int,
        loss_fn: Callable,
        obj_method: str,
        opt_method: str,
        uq_objectives: Union[List[Callable], Tuple[Callable, ...]],
        nn_kwargs: Dict,
        k_kwargs: Dict,
        opt_kwargs: Dict,
        verbose: bool = False,
    ) -> None:
        (
            muygps,
            surrogate_predictions,
            predicted_labels,
            masks,
            acc,
        ) = self._do_classify_uq(
            train,
            test,
            nn_count,
            opt_batch_count,
            uq_batch_count,
            loss_fn,
            obj_method,
            opt_method,
            uq_objectives,
            nn_kwargs,
            k_kwargs,
            opt_kwargs,
            verbose=verbose,
        )
        self.assertEqual(surrogate_predictions.shape, test["output"].shape)
        self.assertEqual(predicted_labels.shape, (test["output"].shape[0],))
        self.assertSequenceAlmostEqual(
            np.sum(surrogate_predictions, axis=1),
            np.zeros(surrogate_predictions.shape[0]),
        )
        # There is almost certainly a better way to do this.
        if np.all(np.unique(train["output"]) == np.unique([-1, 1])):
            self.assertSequenceAlmostEqual(
                np.unique(predicted_labels),
                np.unique(train["output"]),
            )
        else:
            self.assertSequenceAlmostEqual(
                np.unique(predicted_labels),
                np.unique(np.argmax(train["output"], axis=1)),
            )
        print("Finds hyperparameters:")
        param_names, param_vals, _ = muygps.get_opt_params()
        for i, p in enumerate(param_names):
            print(f"\t{p} : {param_vals[i]}")
        print(f"obtains accuracy: {acc}")
        self.assertGreaterEqual(acc, target_acc)
        accuracy, uq = do_uq(surrogate_predictions, test["output"], masks)
        self.assertEqual(
            masks.shape, (len(uq_objectives), test["output"].shape[0])
        )
        self.assertEqual(uq.shape, (len(uq_objectives), 3))
        # @NOTE[bwp] Should we do more to validate the uq? Expected ranges?
        # What about the first dimension `np.sum(mask)`, which records the
        # number of "ambiguous" prediction locations?
        # print(uq)
        if isinstance(muygps.kernel, Matern):
            for i in range(uq.shape[0] - 1):
                self.assertLessEqual(uq[i, 1], acc)
                self.assertGreaterEqual(uq[i, 2], acc)
        self.assertLessEqual(uq[-1, 1], 1.0)
        self.assertGreaterEqual(uq[-1, 2], target_acc)

    def _do_classify_uq(
        self,
        train: Dict[str, np.ndarray],
        test: Dict[str, np.ndarray],
        nn_count: int,
        opt_batch_count: int,
        uq_batch_count: int,
        loss_fn: Callable,
        obj_method: str,
        opt_method: str,
        uq_objectives: Union[List[Callable], Tuple[Callable, ...]],
        nn_kwargs: Dict,
        k_kwargs: Dict,
        opt_kwargs: Dict,
        verbose: bool = False,
    ) -> Tuple[MuyGPS, np.ndarray, np.ndarray, np.ndarray, float]:
        muygps, _, surrogate_predictions, masks = do_classify_uq(
            test["input"],
            train["input"],
            train["output"],
            nn_count=nn_count,
            opt_batch_count=opt_batch_count,
            uq_batch_count=uq_batch_count,
            loss_fn=loss_fn,
            obj_method=obj_method,
            opt_method=opt_method,
            uq_objectives=uq_objectives,
            k_kwargs=k_kwargs,
            nn_kwargs=nn_kwargs,
            opt_kwargs=opt_kwargs,
            verbose=verbose,
        )

        predicted_labels = np.argmax(surrogate_predictions, axis=1)
        acc = np.mean(predicted_labels == np.argmax(test["output"], axis=1))
        if np.all(np.unique(train["output"]) == np.unique([-1.0, 1.0])):
            predicted_labels = 2 * predicted_labels - 1
        return muygps, surrogate_predictions, predicted_labels, masks, acc


class RegressionAPITest(parameterized.TestCase):
    def _do_regress_test_chassis(
        self,
        train: Dict[str, np.ndarray],
        test: Dict[str, np.ndarray],
        target_mse: float,
        nn_count: int,
        batch_count: int,
        loss_fn: Callable,
        obj_method: str,
        opt_method: str,
        sigma_method: Optional[str],
        nn_kwargs: Dict,
        k_kwargs: Union[Dict, Union[List[Dict], Tuple[Dict, ...]]],
        opt_kwargs: Dict,
        verbose: bool = False,
    ) -> None:
        regressor, predictions, mse, variance = self._do_regress(
            train,
            test,
            nn_count,
            batch_count,
            loss_fn,
            obj_method,
            opt_method,
            sigma_method,
            nn_kwargs,
            k_kwargs,
            opt_kwargs,
            verbose=verbose,
        )
        self.assertEqual(predictions.shape, test["output"].shape)
        if isinstance(regressor, MuyGPS):
            self._verify_regressor(
                regressor, variance, test["output"], sigma_method
            )
        else:
            test_count, _ = test["output"].shape
            for i, model in enumerate(regressor.models):
                self._verify_regressor(
                    model,
                    variance[:, i].reshape(test_count, 1),
                    test["output"][:, i].reshape(test_count, 1),
                    sigma_method,
                )
        print(f"obtains mse: {mse}")
        self.assertLessEqual(mse, target_mse)

    def _verify_regressor(self, regressor, variance, targets, sigma_method):
        param_names, param_vals, _ = regressor.get_opt_params()
        if len(param_names) > 0:
            print("finds hyperparameters:")
            for i, p in enumerate(param_names):
                print(f"\t{p} : {param_vals[i]}")
        if variance is not None:
            test_count, response_count = targets.shape
            self.assertEqual(variance.shape, (test_count, response_count))
        if sigma_method is None:
            self.assertFalse(regressor.sigma_sq.trained)
        elif sigma_method.lower() == "analytic":
            self.assertEqual(regressor.sigma_sq.shape, (response_count,))
            _check_ndarray(self.assertEqual, regressor.sigma_sq(), np.ftype)
        else:
            raise ValueError(f"Unsupported sigma method {sigma_method}")

    def _do_regress(
        self,
        train: Dict[str, np.ndarray],
        test: Dict[str, np.ndarray],
        nn_count: int,
        batch_count: int,
        loss_fn: Callable,
        obj_method: str,
        opt_method: str,
        sigma_method: Optional[str],
        nn_kwargs: Dict,
        k_kwargs: Union[Dict, Union[List[Dict], Tuple[Dict, ...]]],
        opt_kwargs: Dict,
        verbose: bool = False,
    ) -> Tuple[Union[MuyGPS, MMuyGPS], np.ndarray, float, np.ndarray,]:
        # print("gets here")
        regressor, _, predictions, variance = do_regress(
            test["input"],
            train["input"],
            train["output"],
            nn_count=nn_count,
            batch_count=batch_count,
            loss_fn=loss_fn,
            obj_method=obj_method,
            opt_method=opt_method,
            sigma_method=sigma_method,
            k_kwargs=k_kwargs,
            nn_kwargs=nn_kwargs,
            opt_kwargs=opt_kwargs,
            verbose=verbose,
        )

        predictions = _consistent_unchunk_tensor(predictions)
        variance = _consistent_unchunk_tensor(variance)
        mse = mse_fn(predictions, test["output"])
        return (regressor, predictions, mse, variance)


class FastPosteriorMeanAPITest(parameterized.TestCase):
    def _do_fast_posterior_mean_test_chassis(
        self,
        train: Dict[str, np.ndarray],
        test: Dict[str, np.ndarray],
        target_mse: float,
        nn_count: int,
        batch_count: int,
        loss_fn: Callable,
        obj_method: str,
        opt_method: str,
        nn_kwargs: Dict,
        k_kwargs: Union[Dict, Union[List[Dict], Tuple[Dict, ...]]],
        opt_kwargs: Dict,
        verbose: bool = False,
    ) -> None:
        regressor, predictions, mse = self._do_fast_posterior_mean(
            train=train,
            test=test,
            nn_count=nn_count,
            batch_count=batch_count,
            loss_fn=loss_fn,
            obj_method=obj_method,
            opt_method=opt_method,
            nn_kwargs=nn_kwargs,
            k_kwargs=k_kwargs,
            opt_kwargs=opt_kwargs,
            verbose=verbose,
        )
        self.assertEqual(predictions.shape, test["output"].shape)
        print(f"obtains mse: {mse}")
        self.assertLessEqual(mse, target_mse)

        if isinstance(regressor, MuyGPS):
            param_names, param_vals, _ = regressor.get_opt_params()
            for i, p in enumerate(param_names):
                print(f"\t{p} : {param_vals[i]}")
        elif isinstance(regressor, MMuyGPS):
            for i, model in enumerate(regressor.models):
                print(f"model {i}:")
                param_names, param_vals, _ = model.get_opt_params()
                for i, p in enumerate(param_names):
                    print(f"\t{p} : {param_vals[i]}")

    def _do_fast_posterior_mean(
        self,
        train: Dict[str, np.ndarray],
        test: Dict[str, np.ndarray],
        nn_count: int,
        batch_count: int,
        loss_fn: Callable,
        obj_method: str,
        opt_method: str,
        nn_kwargs: Dict,
        k_kwargs: Union[Dict, Union[List[Dict], Tuple[Dict, ...]]],
        opt_kwargs: Dict,
        verbose: bool = False,
    ) -> Tuple[Union[MuyGPS, MMuyGPS], np.ndarray, float]:
        (
            regressor,
            _,
            predictions,
            precomputed_coefficient_matrix,
            _,
        ) = do_fast_posterior_mean(
            test["input"],
            train["input"],
            train["output"],
            nn_count=nn_count,
            batch_count=batch_count,
            loss_fn=loss_fn,
            obj_method=obj_method,
            opt_method=opt_method,
            k_kwargs=k_kwargs,
            nn_kwargs=nn_kwargs,
            opt_kwargs=opt_kwargs,
            verbose=verbose,
        )

        mse = mse_fn(predictions, test["output"])
        return (
            regressor,
            predictions,
            mse,
        )



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT


from typing import Tuple

import MuyGPyS._src.math.numpy as np
from MuyGPyS._src.mpi_utils import (
    _chunk_function_tensor,
)
from MuyGPyS._src.gp.tensors.numpy import (
    _crosswise_tensor as _crosswise_tensor_n,
    _pairwise_tensor as _pairwise_tensor_n,
    _batch_features_tensor as _batch_features_tensor_n,
    _make_train_tensors as _make_train_tensors_n,
    _make_predict_tensors as _make_predict_tensors_n,
    _make_heteroscedastic_tensor as _make_heteroscedatic_tensor_n,
    _F2,
    _l2,
)


def _make_heteroscedastic_tensor(
    measurement_noise: np.ndarray,
    batch_nn_indices: np.ndarray,
) -> np.ndarray:
    return _chunk_function_tensor(
        _make_heteroscedatic_tensor_n,
        measurement_noise,
        batch_nn_indices,
    )


def _make_fast_predict_tensors(
    batch_nn_indices: np.ndarray,
    train_features: np.ndarray,
    train_targets: np.ndarray,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    raise NotImplementedError(
        'Function "make_fast_predict_tensors" does not support mpi!'
    )


def _make_predict_tensors(
    batch_indices: np.ndarray,
    batch_nn_indices: np.ndarray,
    test_features: np.ndarray,
    train_features: np.ndarray,
    train_targets: np.ndarray,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    return _chunk_function_tensor(
        _make_predict_tensors_n,
        batch_indices,
        batch_nn_indices,
        test_features,
        train_features,
        train_targets,
        return_count=3,
    )


def _make_train_tensors(
    batch_indices: np.ndarray,
    batch_nn_indices: np.ndarray,
    train_features: np.ndarray,
    train_targets: np.ndarray,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    return _chunk_function_tensor(
        _make_train_tensors_n,
        batch_indices,
        batch_nn_indices,
        train_features,
        train_targets,
        return_count=4,
    )


def _batch_features_tensor(
    features: np.ndarray,
    batch_indices: np.ndarray,
) -> np.ndarray:
    _batch_features_tensor_n
    return _chunk_function_tensor(
        _batch_features_tensor_n,
        features,
        batch_indices,
    )


def _crosswise_tensor(
    data: np.ndarray,
    nn_data: np.ndarray,
    data_indices: np.ndarray,
    nn_indices: np.ndarray,
) -> np.ndarray:
    return _chunk_function_tensor(
        _crosswise_tensor_n,
        data,
        nn_data,
        data_indices,
        nn_indices,
    )


def _crosswise_differences(
    locations: np.ndarray, points: np.ndarray
) -> np.ndarray:
    raise NotImplementedError(
        'Function "muygps_crosswise_differences" does not support mpi!'
    )


def _pairwise_tensor(
    data: np.ndarray,
    nn_indices: np.ndarray,
) -> np.ndarray:
    return _chunk_function_tensor(
        _pairwise_tensor_n,
        data,
        nn_indices,
    )


def _pairwise_differences(points: np.ndarray) -> np.ndarray:
    raise NotImplementedError(
        'Function "muygps_pairwise_differences" does not support mpi!'
    )


def _fast_nn_update(
    train_nn_indices: np.ndarray,
) -> np.ndarray:
    raise NotImplementedError(
        'Function "muygps_fast_nn_update" does not support mpi!'
    )



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from numpy import (
    all,
    allclose,
    any,
    argmax,
    argmin,
    atleast_1d,
    atleast_2d,
    concatenate,
    copy,
    corrcoef,
    cov,
    count_nonzero,
    choose,
    divide,
    einsum,
    exp,
    expand_dims,
    finfo,
    float32,
    float64,
    histogram,
    inf,
    int32,
    int64,
    isclose,
    isnan,
    linalg,
    log,
    logical_and,
    logical_or,
    max,
    mean,
    median,
    meshgrid,
    min,
    mod,
    nan,
    ndarray,
    number,
    outer,
    prod,
    random,
    repeat,
    reshape,
    sqrt,
    sum,
    tile,
    unique,
    where,
    vstack,
)
from numpy import (
    arange as _arange,
    array as _array,
    diagonal as _diagonal,
    eye as _eye,
    full as _full,
    linspace as _linspace,
    ones as _ones,
    zeros as _zeros,
)

# from numpy import *
from numpy.linalg import cholesky

from MuyGPyS._src.math.meta import (
    fix_function_type,
    fix_function_types,
    set_type,
)


def assign(x: ndarray, y: ndarray, *slices) -> ndarray:
    ret = copy(x)
    ret[slices] = y
    return ret


ftype = set_type(float64, float32)
# itype = set_type(int64, int32)
itype = int64

iarray = fix_function_type(itype, _array)
farray = fix_function_type(ftype, _array)
array = farray

arange = fix_function_type(itype, _arange)
diagonal, eye, full, linspace, ones, zeros = fix_function_types(
    ftype, _diagonal, _eye, _full, _linspace, _ones, _zeros
)


def parameter(x):
    return x



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

"""
Objective Handling

MuyGPyS includes predefined objective functions and convenience functions for
indicating them to optimization.
"""

from inspect import signature
from typing import Callable, Dict, Optional

import MuyGPyS._src.math as mm


def make_obj_fn(obj_method: str, *args, **kwargs) -> Callable:
    """
    Prepare an objective function as a function purely of the hyperparameters
    to be optimized.

    This function is designed for use with
    :func:`MuyGPyS.optimize.chassis.optimize_from_tensors()`, and the format
    depends on the `opt_method` argument.

    Args:
        obj_method:
            The name of the objective function to be minimized.
        opt_method:
            The name of the optimization method to be utilized.

    Returns:
        A Callable `objective_fn`, whose format depends on `opt_method`.
    """
    if obj_method == "loo_crossval":
        return make_loo_crossval_fn(*args, **kwargs)
    else:
        raise ValueError(f"Unsupported objective method: {obj_method}")


def make_loo_crossval_fn(
    loss_fn: Callable,
    kernel_fn: Callable,
    mean_fn: Callable,
    var_fn: Callable,
    sigma_sq_fn: Callable,
    pairwise_diffs: mm.ndarray,
    crosswise_diffs: mm.ndarray,
    batch_nn_targets: mm.ndarray,
    batch_targets: mm.ndarray,
    batch_features: Optional[mm.ndarray] = None,
    loss_kwargs: Dict = dict(),
) -> Callable:
    """
    Prepare a leave-one-out cross validation function as a function purely of
    the hyperparameters to be optimized.

    This function is designed for use with
    :func:`MuyGPyS.optimize.chassis.optimize_from_tensors()`, and the format
    depends on the `opt_method` argument.

    Args:
        loss_fn:
            The loss functor used to evaluate model performance.
        kernel_fn:
            A function that realizes kernel tensors given a list of the free
            parameters.
        mean_fn:
            A function that realizes MuyGPs posterior mean prediction given an
            epsilon value. The given value is unused if epsilon is fixed.
        var_fn:
            A function that realizes MuyGPs posterior variance prediction given
            an epsilon value. The given value is unused if epsilon is fixed.
        sigma_sq_fn:
            A function that realizes `sigma_sq` optimization given an epsilon
            value. The given value is unused if epsilon is fixed.
        pairwise_diffs:
            A tensor of shape `(batch_count, nn_count, nn_count, feature_count)`
            containing the `(nn_count, nn_count, feature_count)`-shaped pairwise
            nearest neighbor difference tensors corresponding to each of the
            batch elements.
        crosswise_diffs:
            A tensor of shape `(batch_count, nn_count, feature_count)` whose
            last two dimensions list the difference between each feature of each
            batch element element and its nearest neighbors.
        batch_nn_targets:
            Tensor of floats of shape `(batch_count, nn_count, response_count)`
            containing the expected response for each nearest neighbor of each
            batch element.
        batch_targets:
            Matrix of floats of shape `(batch_count, response_count)` whose rows
            give the expected response for each  batch element.
        batch_features:
            Matrix of floats of shape `(batch_count, feature_count)` whose rows
            give the features for each batch element.
        loss_kwargs:
            A dict listing any additional kwargs to pass to the loss function.

    Returns:
        A Callable `objective_fn`, whose format depends on `opt_method`.
    """
    kernels_fn = make_kernels_fn(kernel_fn, pairwise_diffs, crosswise_diffs)
    # This is ad-hoc, and might need to be revisited.
    if "variances" in signature(loss_fn).parameters.keys():
        predict_and_loss_fn = make_var_predict_and_loss_fn(
            loss_fn,
            mean_fn,
            var_fn,
            sigma_sq_fn,
            batch_nn_targets,
            batch_targets,
            **loss_kwargs,
        )
    else:
        predict_and_loss_fn = make_raw_predict_and_loss_fn(
            loss_fn,
            mean_fn,
            var_fn,
            sigma_sq_fn,
            batch_nn_targets,
            batch_targets,
            **loss_kwargs,
        )

    def obj_fn(*args, **kwargs):
        K, Kcross = kernels_fn(*args, batch_features=batch_features, **kwargs)
        return predict_and_loss_fn(K, Kcross, *args, **kwargs)

    return obj_fn


def make_kernels_fn(
    kernel_fn: Callable,
    pairwise_diffs: mm.ndarray,
    crosswise_diffs: mm.ndarray,
) -> Callable:
    def kernels_fn(*args, **kwargs):
        K = kernel_fn(pairwise_diffs, *args, **kwargs)
        Kcross = kernel_fn(crosswise_diffs, *args, **kwargs)
        return K, Kcross

    return kernels_fn


def make_raw_predict_and_loss_fn(
    loss_fn: Callable,
    mean_fn: Callable,
    var_fn: Callable,
    sigma_sq_fn: Callable,
    batch_nn_targets: mm.ndarray,
    batch_targets: mm.ndarray,
    **loss_kwargs,
) -> Callable:
    def predict_and_loss_fn(K, Kcross, *args, **kwargs):
        predictions = mean_fn(
            K,
            Kcross,
            batch_nn_targets,
            **kwargs,
        )

        return -loss_fn(predictions, batch_targets, **loss_kwargs)

    return predict_and_loss_fn


def make_var_predict_and_loss_fn(
    loss_fn: Callable,
    mean_fn: Callable,
    var_fn: Callable,
    sigma_sq_fn: Callable,
    batch_nn_targets: mm.ndarray,
    batch_targets: mm.ndarray,
    **loss_kwargs,
) -> Callable:
    def predict_and_loss_fn(K, Kcross, *args, **kwargs):
        predictions = mean_fn(
            K,
            Kcross,
            batch_nn_targets,
            **kwargs,
        )
        sigma_sq = sigma_sq_fn(K, batch_nn_targets, **kwargs)

        variances = var_fn(K, Kcross, **kwargs)

        return -loss_fn(
            predictions, batch_targets, variances, sigma_sq, **loss_kwargs
        )

    return predict_and_loss_fn



# Copyright 2023-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from .scalar import (
    _init_scalar_hyperparameter,
    ScalarHyperparameter,
)

from .tensor import (
    TensorHyperparameter,
    _init_tensor_hyperparameter,
)



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

try:
    from jax import config as jax_config
    from jax._src.config import Config as JaxConfig
except Exception:
    jax_config = None  # type: ignore
    from MuyGPyS._src.jaxconfig import Config as JaxConfig  # type: ignore

import itertools
import sys


class MuyGPySConfig(JaxConfig):
    def __init__(self):
        super(MuyGPySConfig, self).__init__()
        self.state = MuyGPySState()
        self.mpi_state = MPIState()

    def parse_flags_with_absl(self):
        if self.state.already_configured_with_absl is False:
            # Extract just the --muygpys... flags (before the first --) from
            # argv. In some environments (e.g. ipython/colab) argv might be a
            # mess of things parseable by absl and other junk.
            muygpys_argv = itertools.takewhile(lambda a: a != "--", sys.argv)
            muygpys_argv = [
                "",
                *(a for a in muygpys_argv if a.startswith("--muygpys")),
            ]

            import absl.flags

            self.config_with_absl()
            absl.flags.FLAGS(muygpys_argv, known_only=True)
            self.complete_absl_config(absl.flags)
            self.state.already_configured_with_absl = True


class MuyGPySState:
    def __init__(self):
        self.jax_enabled = False
        self.torch_enabled = False
        self.hnswlib_enabled = False
        self.gpu_enabled = False
        self.mpi_enabled = False
        self.backend = "numpy"
        self.ftype = "64"
        self.already_configured_with_absl = False

    def low_precision(self):
        return True if self.ftype == "32" else False


class MPIState:
    def __init__(self):
        self._comm_world = None

    @property
    def comm_world(self):
        return self._comm_world

    def set_comm(self, comm):
        self._comm_world = comm


config = MuyGPySConfig()


def _update_jax_global(val):
    config.state.jax_enabled = val


def _update_jax_thread_local(val):
    config.state.jax_enabled = val


# JAX and GPU states

enable_jax = config.define_bool_state(
    name="muygpys_jax_enabled",
    default=False,
    help="Enable use of jax implementations of math functions.",
    update_global_hook=_update_jax_global,
    update_thread_local_hook=_update_jax_thread_local,
)


def _update_gpu_global(val):
    config.state.gpu_enabled = val


def _update_gpu_thread_local(val):
    config.state.gpu_enabled = val


enable_gpu = config.define_bool_state(
    name="muygpys_gpu_enabled",
    default=False,
    help="Enable use of GPUs with JAX.",
    update_global_hook=_update_gpu_global,
    update_thread_local_hook=_update_gpu_thread_local,
)


try:
    from jax import default_backend as _default_backend

    config.update("muygpys_jax_enabled", True)
    if _default_backend() in ["gpu", "tpu"]:
        config.update("muygpys_gpu_enabled", True)
    del _default_backend
except Exception:
    config.update("muygpys_jax_enabled", False)
    config.update("muygpys_gpu_enabled", False)


# hnswlib states


def _update_hnswlib_global(val):
    config.state.hnswlib_enabled = val


def _update_hnswlib_thread_local(val):
    config.state.hnswlib_enabled = val


enable_hnswlib = config.define_bool_state(
    name="muygpys_hnswlib_enabled",
    default=False,
    help="Enable use of hnswlib implementation of fast approximate nearest neighbors.",
    update_global_hook=_update_hnswlib_global,
    update_thread_local_hook=_update_hnswlib_thread_local,
)


try:
    import hnswlib as _hnswlib

    config.update("muygpys_hnswlib_enabled", True)
    del _hnswlib
except Exception:
    config.update("muygpys_hnswlib_enabled", False)


# MPI states


def _update_mpi_global(val):
    config.state.mpi_enabled = val


def _update_mpi_thread_local(val):
    config.state.mpi_enabled = val


enable_mpi = config.define_bool_state(
    name="muygpys_mpi_enabled",
    default=False,
    help="Enable use of mpi for parallelization.",
    update_global_hook=_update_mpi_global,
    update_thread_local_hook=_update_mpi_thread_local,
)

try:
    from mpi4py import MPI
    from mpi4py.util.pkl5 import Intracomm

    # wrap COMM_WORLD with pkl5 for large number of messages per
    # https://mpi4py.readthedocs.io/en/3.1.4/mpi4py.util.pkl5.html
    config.mpi_state.set_comm(Intracomm(MPI.COMM_WORLD))

    config.update("muygpys_mpi_enabled", True)
except Exception:
    MPI = None  # type: ignore
    config.update("muygpys_mpi_enabled", False)


# torch states


def _update_torch_global(val):
    config.state.torch_enabled = val


def _update_torch_thread_local(val):
    config.state.torch_enabled = val


enable_torch = config.define_bool_state(
    name="muygpys_torch_enabled",
    default=False,
    help="Enable use of torch for deep kernel learning.",
    update_global_hook=_update_torch_global,
    update_thread_local_hook=_update_torch_thread_local,
)


try:
    import torch as _torch

    config.update("muygpys_torch_enabled", True)
    del _torch
except Exception:
    config.update("muygpys_torch_enabled", False)

# Backend state


def _update_backend_global(val):
    config.state.backend = val


def _update_backend_thread_local(val):
    config.state.backend = val


backend = config.define_enum_state(
    name="muygpys_backend",
    enum_values=["numpy", "jax", "torch", "mpi"],
    default="numpy",
    help="Specify which backend to select at import time",
    update_global_hook=_update_backend_global,
    update_thread_local_hook=_update_backend_thread_local,
)

if (
    config.state.backend == "jax"
    and config.state.jax_enabled is False
    or config.state.backend == "torch"
    and config.state.torch_enabled is False
    or config.state.backend == "mpi"
    and config.state.mpi_enabled is False
):
    be = config.state.backend
    raise ValueError(
        f'MuyGPyS backend is set to "{be}", but "{be}" is not enabled! '
        f"The {be} dependencies are most likely not installed in your "
        f"environment."
    )


def _update_ftype_global(val):
    config.state.ftype = val


def _update_ftype_thread_local(val):
    config.state.ftype = val


ftype = config.define_enum_state(
    name="muygpys_ftype",
    enum_values=["32", "64"],
    default="64",
    help="Specify the float precision to be used",
    update_global_hook=_update_ftype_global,
    update_thread_local_hook=_update_ftype_thread_local,
)

if config.state.jax_enabled is True and config.state.ftype == "64":
    jax_config.update("jax_enable_x64", True)



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

"""
Convenience functions for optimizing :class:`~MuyGPyS.gp.muygps.MuyGPS` objects

The functions
:func:`~MuyGPyS.optimize.chassis.optimize_from_indices` and
:func:`~MuyGPyS.optimize.chassis.optimize_from_tensors` wrap different
optimization packages to provide a simple interface to optimize the
hyperparameters of :class:`~MuyGPyS.gp.muygps.MuyGPS` objects.

Currently, `opt_method="scipy"` wraps :class:`scipy.optimize.opt`
multiparameter optimization using L-BFGS-B algorithm using the objective
function :func:`MuyGPyS.optimize.objective.loo_crossval`.

Currently, `opt_method="bayesian"` (also accepts `"bayes"` and `"bayes_opt"`)
wraps :class:`bayes_opt.BayesianOptimization`. Unlike the `scipy` version,
`BayesianOptimization` can be meaningfully modified by several kwargs.
`MuyGPyS` assigns reasonable defaults if no settings are passed by the user.
See the `BayesianOptimization <https://github.com/fmfn/BayesianOptimization>`_
documentation for details.
"""


from typing import Callable, Dict, Optional

import MuyGPyS._src.math as mm
from MuyGPyS._src.optimize.chassis import (
    _scipy_optimize,
    _bayes_opt_optimize,
)
from MuyGPyS.gp import MuyGPS
from MuyGPyS.optimize.utils import _switch_on_opt_method
from MuyGPyS.optimize.objective import make_obj_fn
from MuyGPyS.optimize.loss import lool_fn
from MuyGPyS.optimize.sigma_sq import make_sigma_sq_optim


def optimize_from_tensors(
    muygps: MuyGPS,
    batch_targets: mm.ndarray,
    batch_nn_targets: mm.ndarray,
    crosswise_diffs: mm.ndarray,
    pairwise_diffs: mm.ndarray,
    batch_features: Optional[mm.ndarray] = None,
    loss_fn: Callable = lool_fn,
    obj_method: str = "loo_crossval",
    opt_method: str = "bayes",
    sigma_method: Optional[str] = "analytic",
    loss_kwargs: Dict = dict(),
    verbose: bool = False,
    **kwargs,
) -> MuyGPS:
    """
    Find the optimal model using existing difference matrices.

    See the following example, where we have already created a `batch_indices`
    vector and a `batch_nn_indices` matrix using
    :class:`MuyGPyS.neighbors.NN_Wrapper`, a `crosswise_diffs`
    matrix using :func:`MuyGPyS.gp.tensors.crosswise_tensor` and
    `pairwise_diffs` using :func:`MuyGPyS.gp.tensors.pairwise_tensor`, and
    initialized a :class:`~MuyGPyS.gp.muygps.MuyGPS` model `muygps`.

    Example:
        >>> from MuyGPyS.optimize.chassis import optimize_from_tensors
        >>> muygps = optimize_from_tensors(
        ...         muygps,
        ...         batch_indices,
        ...         batch_nn_indices,
        ...         crosswise_diffs,
        ...         pairwise_diffs,
        ...         train_responses,
        ...         loss_method='mse',
        ...         obj_method='loo_crossval',
        ...         opt_method='scipy',
        ...         verbose=True,
        ... )
        parameters to be optimized: ['nu']
        bounds: [[0.1 1. ]]
        sampled x0: [0.8858425]
        optimizer results:
              fun: 0.4797763813693626
         hess_inv: <1x1 LbfgsInvHessProduct with dtype=float64>
              jac: array([-3.06976666e-06])
          message: b'CONVERGENCE: NORM_OF_PROJECTED_GRADIENT_<=_PGTOL'
             nfev: 16
              nit: 5
             njev: 8
           status: 0
          success: True
                x: array([0.39963594])

    Args:
        muygps:
            The model to be optimized.
        batch_targets:
            Matrix of floats of shape `(batch_count, response_count)` whose rows
            give the expected response for each batch element.
        batch_nn_targets:
            Tensor of floats of shape `(batch_count, nn_count, response_count)`
            containing the expected response for each nearest neighbor of each
            batch element.
        crosswise_diffs:
            A tensor of shape `(batch_count, nn_count, feature_count)` whose
            last two dimensions list the difference between each feature of each
            batch element element and its nearest neighbors.
        pairwise_diffs:
            A tensor of shape `(batch_count, nn_count, nn_count, feature_count)`
            containing the `(nn_count, nn_count, feature_count)`-shaped pairwise
            nearest neighbor difference tensors corresponding to each of the
            batch elements.
        loss_fn:
            The loss functor used to evaluate model performance.
        obj_method:
            Indicates the objective function to be minimized. Currently
            restricted to `"loo_crossval"`.
        opt_method:
            Indicates the optimization method to be used. Currently restricted
            to `"bayesian"` (alternately `"bayes"` or `"bayes_opt"`) and
            `"scipy"`.
        sigma_method:
            The optimization method to be employed to learn the `sigma_sq`
            hyperparameter.
        loss_kwargs:
            A dictionary of additional keyword arguments to apply to the loss
            function. Loss function specific.
        verbose:
            If True, print debug messages.
        kwargs:
            Additional keyword arguments to be passed to the wrapper optimizer.

    Returns:
        A new MuyGPs model whose specified hyperparameters have been optimized.
    """
    kernel_fn = muygps.kernel.get_opt_fn()
    mean_fn = muygps.get_opt_mean_fn()
    var_fn = muygps.get_opt_var_fn()
    sigma_sq_fn = make_sigma_sq_optim(sigma_method, muygps)

    obj_fn = make_obj_fn(
        obj_method,
        loss_fn,
        kernel_fn,
        mean_fn,
        var_fn,
        sigma_sq_fn,
        pairwise_diffs,
        crosswise_diffs,
        batch_nn_targets,
        batch_targets,
        batch_features=batch_features,
        loss_kwargs=loss_kwargs,
    )

    return _switch_on_opt_method(
        opt_method,
        _bayes_opt_optimize,
        _scipy_optimize,
        muygps,
        obj_fn,
        verbose=verbose,
        **kwargs,
    )



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from jax import jit
from jax.nn import softmax

import MuyGPyS._src.math.jax as jnp


@jit
def _cross_entropy_fn(
    predictions: jnp.ndarray,
    targets: jnp.ndarray,
    ll_eps: float = 1e-15,
) -> float:
    one_hot_targets = jnp.where(targets > 0.0, 1.0, 0.0)
    softmax_predictions = softmax(predictions, axis=1)

    return _log_loss(one_hot_targets, softmax_predictions, eps=ll_eps)


@jit
def _log_loss(
    y_true: jnp.ndarray, y_pred: jnp.ndarray, eps: float = 1e-15
) -> jnp.ndarray:
    """
    Lifted whole-cloth from [0].

    References
    ----------
    https://github.com/scikit-learn/scikit-learn/blob/7e1e6d09b/sklearn/metrics/_classification.py#L2309
    """
    # Clipping
    y_pred = jnp.clip(y_pred, eps, 1 - eps)

    # Renormalize
    y_pred /= y_pred.sum(axis=1)[:, jnp.newaxis]
    loss = -(y_true * jnp.log(y_pred)).sum(axis=1)

    return jnp.sum(loss)


@jit
def _mse_fn_unnormalized(
    predictions: jnp.ndarray,
    targets: jnp.ndarray,
) -> jnp.ndarray:
    return jnp.sum((predictions - targets) ** 2)


@jit
def _mse_fn(
    predictions: jnp.ndarray,
    targets: jnp.ndarray,
) -> float:
    batch_count, response_count = predictions.shape
    return _mse_fn_unnormalized(predictions, targets) / (
        batch_count * response_count
    )


@jit
def _lool_fn(
    predictions: jnp.ndarray,
    targets: jnp.ndarray,
    variances: jnp.ndarray,
    sigma_sq: jnp.ndarray,
) -> jnp.ndarray:
    return _lool_fn_unscaled(
        predictions, targets, jnp.outer(variances, sigma_sq)
    )


@jit
def _lool_fn_unscaled(
    predictions: jnp.ndarray, targets: jnp.ndarray, variances: jnp.ndarray
) -> float:
    return jnp.sum(
        jnp.divide((predictions - targets) ** 2, variances) + jnp.log(variances)
    )


@jit
def _pseudo_huber_fn(
    predictions: jnp.ndarray, targets: jnp.ndarray, boundary_scale: float = 1.5
) -> float:
    return boundary_scale**2 * jnp.sum(
        jnp.sqrt(1 + jnp.divide(targets - predictions, boundary_scale) ** 2) - 1
    )


@jit
def _looph_fn_unscaled(
    predictions: jnp.ndarray,
    targets: jnp.ndarray,
    variances: jnp.ndarray,
    boundary_scale: float = 1.5,
) -> float:
    boundary_scale_sq = boundary_scale**2
    return jnp.sum(
        boundary_scale_sq
        * jnp.sqrt(
            1
            + jnp.divide(targets - predictions, variances * boundary_scale) ** 2
        )
        - boundary_scale_sq
        + jnp.log(variances)
    )


@jit
def _looph_fn(
    predictions: jnp.ndarray,
    targets: jnp.ndarray,
    variances: jnp.ndarray,
    sigma_sq: jnp.ndarray,
    boundary_scale: float = 1.5,
) -> float:
    return _looph_fn_unscaled(
        predictions,
        targets,
        jnp.outer(variances, sigma_sq),
        boundary_scale=boundary_scale,
    )



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

"""
Hyperparameters and kernel functors

Defines kernel functors (inheriting
:class:`~MuyGPyS.gp.kernels.kernel_fn.KernelFn`) that transform crosswise
difference tensors into cross-covariance matrices and pairwise difference
matrices into covariance or kernel tensors.

See the following example to initialize an :class:`MuyGPyS.gp.kernels.Matern`
object. Other kernel functors are similar, but require different
hyperparameters.

Example:
    >>> from MuyGPyS.gp.kernels import Matern
    >>> kern = Matern(
    ...     nu=ScalarHyperparameter("log_sample", (0.1, 2.5)),
    ...     metric=IsotropicDistortion(
    ...         l2,
    ...         length_scale=ScalarHyperparameter(1.0),
    ...     ),
    ... )

One uses a previously computed `pairwise_diffs` tensor (see
:func:`MuyGPyS.gp.tensor.pairwise_tensor`) to compute a kernel tensor whose
second two dimensions contain square kernel matrices. Similarly, one uses a
previously computed `crosswise_diffs` matrix (see
:func:`MuyGPyS.gp.tensor.crosswise_diffs`) to compute a cross-covariance
matrix. See the following example, which assumes that you have already
constructed the difference `numpy.ndarrays` and the kernel `kern` as shown
above.

Example:
    >>> K = kern(pairwise_diffs)
    >>> Kcross = kern(crosswise_diffs)
"""

from typing import Callable, Dict, List, Tuple, Union

import MuyGPyS._src.math as mm
from MuyGPyS.gp.distortion import (
    AnisotropicDistortion,
    IsotropicDistortion,
    NullDistortion,
)


class KernelFn:
    """
    A kernel functor.

    Base class for kernel functors that include a hyperparameter Dict and a
    call mechanism.

    Args:
        kwargs:
            Ignored (by this base class) keyword arguments.
    """

    def __init__(
        self,
        metric: Union[
            AnisotropicDistortion, IsotropicDistortion, NullDistortion
        ],
    ):
        """
        Initialize dict holding hyperparameters.
        """
        self._hyperparameters: Dict = dict()
        self.distortion_fn = metric
        self._make_base()

    def _make_base(self):
        self.distortion_fn.populate_length_scale(self._hyperparameters)

    def set_params(self, **kwargs) -> None:
        """
        Reset hyperparameters using hyperparameter dict(s).

        Args:
            kwargs:
                Hyperparameter kwargs.
        """
        for name in kwargs:
            self._hyperparameters[name]._set(kwargs[name])

    def __call__(self, diffs: mm.ndarray) -> mm.ndarray:
        raise NotImplementedError(
            "__call__ is not implemented for base KernelFn"
        )

    def get_opt_params(
        self,
    ) -> Tuple[List[str], List[float], List[Tuple[float, float]]]:
        """
        Report lists of unfixed hyperparameter names, values, and bounds.

        Returns
        -------
            names:
                A list of unfixed hyperparameter names.
            params:
                A list of unfixed hyperparameter values.
            bounds:
                A list of unfixed hyperparameter bound tuples.
        """
        return self.distortion_fn.get_opt_params()

    @staticmethod
    def _get_opt_fn(
        kernel_fn: Callable,
        distortion_fn: Union[
            AnisotropicDistortion, IsotropicDistortion, NullDistortion
        ],
    ) -> Callable:
        return distortion_fn.get_opt_fn(kernel_fn)

    def __str__(self) -> str:
        """
        Print state of hyperparameter dict.

        Intended only for testing purposes.
        """
        ret = ""
        for p in self._hyperparameters:
            param = self._hyperparameters[p]
            ret += f"{p} : {param()} - {param.get_bounds()}\n"
        return ret[:-1]



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from typing import Callable, Optional


def _switch_on_opt_method(
    opt_method: str, bayes_func: Callable, scipy_func: Callable, *args, **kwargs
):
    opt_method = opt_method.lower()
    if opt_method in ["bayesian", "bayes", "bayes-opt"]:
        return bayes_func(*args, **kwargs)
    elif opt_method == "scipy":
        return scipy_func(*args, **kwargs)
    else:
        raise ValueError(f"Unsupported optimization method: {opt_method}")


def _switch_on_loss_method(
    loss_method: str,
    cross_entropy_func: Callable,
    mse_func: Callable,
    lool_func: Callable,
    pseudo_huber_func: Callable,
    looph_func: Callable,
    *args,
    **kwargs,
):
    loss_method = loss_method.lower()
    if loss_method in ["cross-entropy", "log"]:
        return cross_entropy_func(*args, **kwargs)
    elif loss_method == "mse":
        return mse_func(*args, **kwargs)
    elif loss_method == "lool":
        return lool_func(*args, **kwargs)
    elif loss_method in ["huber", "pseudo-huber", "pseudo_huber"]:
        return pseudo_huber_func(*args, **kwargs)
    elif loss_method in ["looph"]:
        return looph_func(*args, **kwargs)
    else:
        raise NotImplementedError(
            f"Loss function {loss_method} is not implemented."
        )


def _switch_on_sigma_method(
    sigma_method: Optional[str],
    analytic_func: Callable,
    none_func: Callable,
    *args,
    **kwargs,
):
    if sigma_method is not None:
        sigma_method = sigma_method.lower()
    if sigma_method == "analytic":
        return analytic_func(*args, **kwargs)
    elif sigma_method is None:
        return none_func(*args, **kwargs)
    else:
        raise ValueError(f"Unrecognized sigma_method {sigma_method}")



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

import MuyGPyS._src.math.torch as torch


def _analytic_sigma_sq_optim_unnormalized(
    K: torch.ndarray,
    nn_targets: torch.ndarray,
) -> torch.ndarray:
    return torch.sum(
        torch.einsum(
            "ijk,ijk->ik", nn_targets, torch.linalg.solve(K, nn_targets)
        ),
        axis=0,
    )


def _analytic_sigma_sq_optim(
    K: torch.ndarray,
    nn_targets: torch.ndarray,
) -> torch.ndarray:
    batch_count, nn_count, _ = nn_targets.shape
    return _analytic_sigma_sq_optim_unnormalized(K, nn_targets) / (
        nn_count * batch_count
    )



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from MuyGPyS._src.util import _collect_implementation

# _collect_implementation returns a tuple, so need to subscript to get singleton
_shear_fn = _collect_implementation(
    "MuyGPyS._src.gp.kernels.shear", "_shear_fn"
)[0]



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from MuyGPyS._src.util import _collect_implementation

[_analytic_sigma_sq_optim] = _collect_implementation(
    "MuyGPyS._src.optimize.sigma_sq",
    "_analytic_sigma_sq_optim",
)



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from MuyGPyS._src.util import _collect_implementation

(_homoscedastic_perturb, _heteroscedastic_perturb) = _collect_implementation(
    "MuyGPyS._src.gp.noise",
    "_homoscedastic_perturb",
    "_heteroscedastic_perturb",
)



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

"""Public MuyGPyS modules and functions."""

__version__ = "0.7.2"

from MuyGPyS._src.config import (
    config as config,
    jax_config as jax_config,
    MPI as MPI,
)



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

"""
Noise perturbation function wrapper
"""

from typing import Callable, Union

from MuyGPyS._src.gp.noise import (
    _homoscedastic_perturb,
    _heteroscedastic_perturb,
)
from MuyGPyS.gp.noise.homoscedastic import HomoscedasticNoise
from MuyGPyS.gp.noise.heteroscedastic import HeteroscedasticNoise
from MuyGPyS.gp.noise.null import NullNoise


def select_perturb_fn(
    eps: Union[HomoscedasticNoise, HeteroscedasticNoise, NullNoise]
) -> Callable:
    if isinstance(eps, HomoscedasticNoise):
        return _homoscedastic_perturb
    elif isinstance(eps, HeteroscedasticNoise):
        return _heteroscedastic_perturb
    elif isinstance(eps, NullNoise):

        def _null_noise(K, *args, **kwargs):
            return K

        return _null_noise

    else:
        raise ValueError(f"Noise model {type(eps)} is not supported")


def noise_perturb(perturb_fn: Callable):
    def perturbed_version(fn):
        def fn_wrapper(K, *args, eps=0.0, **kwargs):
            return fn(perturb_fn(K, eps), *args, **kwargs)

        return fn_wrapper

    return perturbed_version


def perturb_with_noise_model(
    fn: Callable,
    eps: Union[HomoscedasticNoise, HeteroscedasticNoise, NullNoise],
) -> Callable:
    return noise_perturb(select_perturb_fn(eps))(fn)



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

"""
MuyGPs PyTorch implementation
"""
import MuyGPyS._src.math.torch as torch
from MuyGPyS import config
from MuyGPyS._src.math.torch import nn
from MuyGPyS.gp.tensors import (
    pairwise_tensor,
    crosswise_tensor,
)

from MuyGPyS.gp.multivariate_muygps import MultivariateMuyGPS as MMuyGPS

if config.state.backend != "torch":
    import warnings

    warnings.warn(
        f"torch-only code cannot be run in {config.state.backend} mode"
    )


class MultivariateMuyGPs_layer(nn.Module):
    """
    Multivariate MuyGPs model written as a custom PyTorch layer using nn.Module.

    Implements the MuyGPs algorithm as articulated in [muyskens2021muygps]_. See
    documentation on MuyGPs class for more detail.

    The MultivariateMuyGPs_layer class only supports the RBF and Matern kernels
    currently. More kernels will be added to the torch module of MuyGPs in
    future releases.

    PyTorch does not currently support the Bessel function required to compute
    the Matern kernel for non-special values of :math:`\\nu`, e.g. 1/2, 3/2,
    5/2, and :math:`\\infty`. The MuyGPs layer allows the lengthscale parameter
    :math:`\\rho` to be trained (provided an initial value by the user) as well
    as the homoskedastic :math:`\\varepsilon` noise parameter.

    The MultivariateMuyGPs layer returns the posterior mean, posterior variance
    , and a vector of :math:`\\sigma^2` indicating the scale parameter
    associated with the posterior variance of each dimension of the response.

    :math:`\\sigma^2` is the only parameter assumed to be a training target by
    default, and is treated differently from all other hyperparameters. All
    other training targets must be manually specified in the construction of
    a MuyGPs_layer object.

    Example:
        >>> from MuyGPyS.torch.muygps_layer import MultivariateMuyGPs_layer
        >>> multivariate_muygps_model = MMuyGPs(
        ...     Matern(
        ...         nu=ScalarHyperparameter("sample", (0.1, 1)),
        ...         metric=IsotropicDistortion(
        ...             l2,
        ...             length_scale=ScalarHyperparameter(1.0)
        ...         ),
        ...     ),
        ...     eps=HomoscedasticNoise(1e-5),
        ... )
        >>> batch_indices = torch.arange(100,)
        >>> batch_nn_indices = torch.arange(100,)
        >>> batch_targets = torch.ones(100,)
        >>> batch_nn_targets = torch.ones(100,)
        >>> muygps_layer_object = MultivariateMuyGPs_layer(
        ... multivariate_muygps_model,
        ... batch_indices,
        ... batch_nn_indices,
        ... batch_targets,
        ... batch_nn_targets)



    Args:
        multivariate_muygps_model:
            A MultivariateMuyGPS object to be used in the final GP layer.
        batch_indices:
            A torch.Tensor of shape `(batch_count,)` containing the indices of
            the training data to be sampled for training.
        batch_nn_indices:
            A torch.Tensor of shape `(batch_count, nn_count)` containing the
            indices of the k nearest neighbors of the batched training samples.
        batch_targets:
            A torch.Tensor of shape `(batch_count, response_count)` containing
            the responses corresponding to each batched training sample.
        batch_nn_targets:
            A torch.Tensor of shape `(batch_count, nn_count, response_count)`
            containing the responses corresponding to the nearest neighbors
            of each batched training sample.


        kwargs:
            Addition parameters to be passed to the kernel, possibly including
            additional hyperparameter dicts and a metric keyword.
    """

    def __init__(
        self,
        multivariate_muygps_model: MMuyGPS,
        batch_indices,
        batch_nn_indices,
        batch_targets,
        batch_nn_targets,
    ):
        super().__init__()
        self.multivariate_muygps_model = multivariate_muygps_model
        self.batch_indices = batch_indices
        self.batch_nn_indices = batch_nn_indices
        self.batch_targets = batch_targets
        self.batch_nn_targets = batch_nn_targets

    def forward(self, x):
        """
        Produce the output of a MuyGPs custom PyTorch layer.

        Returns
        -------
        predictions:
            A torch.ndarray of shape `(batch_count, response_count)` whose rows
            are the predicted response for each of the given batch feature.
        variances:
            A torch.ndarray of shape `(batch_count, response_count)`
            consisting of the diagonal elements of the posterior variance.
        """
        crosswise_diffs = crosswise_tensor(
            x,
            x,
            self.batch_indices,
            self.batch_nn_indices,
        )

        pairwise_diffs = pairwise_tensor(x, self.batch_nn_indices)

        batch_count, nn_count, response_count = self.batch_nn_targets.shape

        Kcross = torch.zeros(batch_count, nn_count, response_count)
        K = torch.zeros(batch_count, nn_count, nn_count, response_count)

        for i, model in enumerate(self.multivariate_muygps_model.models):
            Kcross[:, :, i] = model.kernel(crosswise_diffs)
            K[:, :, :, i] = model.kernel(pairwise_diffs)

        variances = self.multivariate_muygps_model.posterior_variance(K, Kcross)

        predictions = self.multivariate_muygps_model.posterior_mean(
            K, Kcross, self.batch_nn_targets
        )

        return predictions, variances



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT
from .multivariate_muygps_layer import MultivariateMuyGPs_layer
from .muygps_layer import MuyGPs_layer



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

"""
Sigma Square hyperparameter
"""

from typing import Callable, Tuple

import MuyGPyS._src.math as mm
from MuyGPyS._src.util import _fullname


class SigmaSq:
    """
    A :math:`\\sigma^2` covariance scale parameter.

    :math:`\\sigma^2` is a scaling parameter that one multiplies with the
    found diagonal variances of a :class:`MuyGPyS.gp.muygps.MuyGPS` or
    :class:`MuyGPyS.gp.muygps.MultivariateMuyGPS` regression in order to obtain
    the predicted posterior variance. Trained values assume a number of
    dimensions equal to the number of response dimensions, and correspond to
    scalar scaling parameters along the corresponding dimensions.

    Args:
        response_count:
            The integer number of response dimensions.
    """

    def __init__(self, response_count: int = 1):
        self.val = mm.ones(response_count)
        self._trained = False

    def __str__(self, **kwargs):
        return f"{type(self).__name__}({self.val})"

    def _set(self, val: mm.ndarray) -> None:
        """
        Value setter.

        Args:
            val:
                The new value of the hyperparameter.
        """
        if not isinstance(val, mm.ndarray):
            raise ValueError(
                f"Expected {_fullname(mm.ndarray)} for SigmaSq value update, "
                f"not {_fullname(val.__class__)}"
            )
        if self.val.shape != val.shape:
            raise ValueError(
                f"Bad attempt to assign SigmaSq of shape {self.val.shape} a "
                f"value of shape {val.shape}"
            )
        if val.dtype != mm.ftype:
            val = mm.farray(val)
        self.val = val
        self._trained = True

    def __call__(self) -> mm.ndarray:
        """
        Value accessor.

        Returns:
            The current value of the hyperparameter.
        """
        return self.val

    @property
    def trained(self) -> bool:
        """
        Report whether the value has been set.

        Returns:
            `True` if trained, `False` otherwise.
        """
        return self._trained

    @property
    def shape(self) -> Tuple[int, ...]:
        """
        Report the shape of the SigmaSq value.

        Returns:
            The shape of the SigmaSq value.
        """
        return self.val.shape


def sigma_sq_scale(fn: Callable) -> Callable:
    def scaled_fn(*args, sigma_sq=[1.0], **kwargs):
        return mm.outer(fn(*args, **kwargs), sigma_sq)

    return scaled_fn


def sigma_sq_apply(fn: Callable, sigma_sq: SigmaSq) -> Callable:
    def scaled_fn(*args, **kwargs):
        return fn(*args, sigma_sq=sigma_sq(), **kwargs)

    return scaled_fn



# Copyright 2023-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from typing import Callable, Union, List, Tuple
import re

from scipy.stats.qmc import LatinHypercube

import MuyGPyS._src.math as mm
from MuyGPyS._src.gp.tensors import (
    _pairwise_differences,
    _crosswise_differences,
)
from MuyGPyS.gp.hyperparameter import ScalarHyperparameter


class HierarchicalNonstationaryHyperparameter:
    """
    A MuyGPs kernel or model hierarchical nonstationary hyperparameter.

    HierarchicalNonstationaryHyperparameter are defined by a set of knots,
    initial guesses for the hyperparameter values at each of those knots,
    and a higher-level GP kernel.

    Args:
        knot_features:
            Tensor of floats of shape `(knot_count, feature_count)`
            containing the feature vectors for each knot.
        knot_values:
            List of scalar hyperparameters of length `knot_count`
            containing the initial values and optimization bounds for each knot.
            Float values will be converted to fixed scalar hyperparameters.
        kernel:
            Initialized higher-level GP kernel.
    """

    def __init__(
        self,
        knot_features: mm.ndarray,
        knot_values: Union[mm.ndarray, List[ScalarHyperparameter]],
        kernel,
    ):
        """
        Initialize a hierarchical nonstationary hyperparameter.
        """
        if len(knot_values) != len(knot_features):
            raise ValueError(
                "knot_features and knot_values must have the same length"
            )
        self._knot_features = knot_features
        self._knot_value_params = [
            knot_value
            if isinstance(knot_value, ScalarHyperparameter)
            else ScalarHyperparameter(float(knot_value))
            for knot_value in knot_values
        ]
        self._kernel = kernel
        self._higher_K = self._kernel(
            _pairwise_differences(self._knot_features)
        )
        self._update_solve()

    def _update_solve(self):
        self._solve = mm.linalg.solve(self._higher_K, self._knot_values())

    def _knot_values(self) -> mm.ndarray:
        return mm.array([[param()] for param in self._knot_value_params])

    @staticmethod
    def _get_knot_key(name: str, index: int) -> str:
        return f"{name}_knot{index}"

    def _update_knot_values(self, **kwargs):
        updated = False
        pattern = re.compile(self._get_knot_key(name=r"(\w+)", index=r"(\d+)"))
        for key, arg in kwargs.items():
            match = pattern.fullmatch(key)
            if match:
                name = match.group(1)
                if name == self._name:
                    index = int(match.group(2))
                    self._knot_value_params[index]._set_val(arg)
                    updated = True
        if updated:
            self._update_solve()

    def __call__(self, batch_features, **kwargs) -> mm.ndarray:
        """
        Value accessor.

        Returns:
            The current value of the hierarchical nonstationary hyperparameter.
        """
        self._update_knot_values(**kwargs)
        lower_Kcross = self._kernel(
            _crosswise_differences(batch_features, self._knot_features)
        )
        return lower_Kcross @ self._solve

    def fixed(self) -> bool:
        """
        Report whether the parameter is fixed, and is to be ignored during optimization.

        Returns:
            `True` if fixed, `False` otherwise.
        """
        return all(param.fixed() for param in self._knot_value_params)

    def get_bounds(self) -> Tuple[float, float]:
        """
        Bounds accessor.

        Returns:
            The lower and upper bound tuple.
        """
        raise NotImplementedError(
            "HierarchicalNonstationaryHyperparameter does not support optimization bounds directly. "
            "Set bounds on individual knot values instead."
        )

    def apply(self, fn: Callable, name: str) -> Callable:
        self._name = name
        if any(param.fixed() for param in self._knot_value_params):

            def applied_fn(*args, **kwargs):
                for index, param in enumerate(self._knot_value_params):
                    if param.fixed():
                        knot_name = self._get_knot_key(name, index)
                        kwargs.setdefault(knot_name, param())
                return fn(*args, **kwargs)

            return applied_fn

        return fn

    def append_lists(
        self,
        name: str,
        names: List[str],
        params: List[float],
        bounds: List[Tuple[float, float]],
    ):
        if not self.fixed():
            for index, param in enumerate(self._knot_value_params):
                knot_name = self._get_knot_key(name, index)
                param.append_lists(knot_name, names, params, bounds)


def sample_knots(feature_count: int, knot_count: int) -> mm.ndarray:
    """
    Samples knots from feature matrix.

    Args:
        feature_count:
            Dimension of feature vectors.
        knot_count:
            Number of knots to sample.

    Returns:
        Tensor of floats of shape `(knot_count, feature_count)`
        containing the sampled feature vectors for each knot.
    """
    return LatinHypercube(feature_count, centered=True).random(knot_count)



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

"""
MuyGPs implementation
"""

from typing import Callable, List, Tuple, Union
from copy import deepcopy

import MuyGPyS._src.math as mm
from MuyGPyS._src.util import auto_str
from MuyGPyS.gp.kernels import (
    Matern,
    RBF,
)
from MuyGPyS.gp.mean import PosteriorMean
from MuyGPyS.gp.sigma_sq import SigmaSq
from MuyGPyS.gp.variance import PosteriorVariance
from MuyGPyS.gp.noise import HeteroscedasticNoise, HomoscedasticNoise, NullNoise
from MuyGPyS.gp.fast_mean import FastPosteriorMean
from MuyGPyS.gp.fast_precompute import FastPrecomputeCoefficients


@auto_str
class MuyGPS:
    """
    Local Kriging Gaussian Process.

    Performs approximate GP inference by locally approximating an observation's
    response using its nearest neighbors. Implements the MuyGPs algorithm as
    articulated in [muyskens2021muygps]_.

    Kernels accept different hyperparameter dictionaries specifying
    hyperparameter settings. Keys can include `val` and `bounds`. `bounds` must
    be either a len == 2  iterable container whose elements are scalars in
    increasing order, or the string `fixed`. If `bounds == fixed` (the default
    behavior), the hyperparameter value will remain fixed during optimization.
    `val` must be either a scalar (within the range of the upper and lower
    bounds if given) or the strings `"sample"` or `log_sample"`, which will
    randomly sample a value within the range given by the bounds.

    In addition to individual kernel hyperparamters, each MuyGPS object also
    possesses a homoscedastic :math:`\\varepsilon` noise parameter and a
    vector of :math:`\\sigma^2` indicating the scale parameter associated
    with the posterior variance of each dimension of the response.

    :math:`\\sigma^2` is the only parameter assumed to be a training target by
    default, and is treated differently from all other hyperparameters. All
    other training targets must be manually specified in `k_kwargs`.

    Example:
        >>> from MuyGPyS.gp import MuyGPS
        >>> k_kwargs = {
        ...         "kern": "rbf",
        ...         "metric": "F2",
        ...         "eps": {"val": 1e-5},
        ...         "nu": {"val": 0.38, "bounds": (0.1, 2.5)},
        ...         "length_scale": {"val": 7.2},
        ... }
        >>> muygps = MuyGPS(**k_kwarg)

    MuyGPyS depends upon linear operations on specially-constructed tensors in
    order to efficiently estimate GP realizations. One can use (see their
    documentation for details) :func:`MuyGPyS.gp.tensors.pairwise_tensor` to
    construct pairwise difference tensors and
    :func:`MuyGPyS.gp.tensors.crosswise_tensor` to produce crosswise diff
    tensors that `MuyGPS` can then use to construct kernel tensors and
    cross-covariance matrices, respectively.

    We can easily realize kernel tensors using a `MuyGPS` object's `kernel`
    functor once we have computed a `pairwise_diffs` tensor and a
    `crosswise_diffs` matrix.

    Example:
        >>> K = muygps.kernel(pairwise_diffs)
        >>> Kcross = muygps.kernel(crosswise_diffs)


    Args:
        kernel:
            The kernel to be used.
        eps:
            A noise model.
        response_count:
            The number of response dimensions.
    """

    def __init__(
        self,
        kernel: Union[Matern, RBF],
        eps: Union[
            NullNoise, HomoscedasticNoise, HeteroscedasticNoise
        ] = HomoscedasticNoise(0.0, "fixed"),
        response_count: int = 1,
    ):
        self.kernel = kernel
        self.sigma_sq = SigmaSq(response_count)
        self.eps = eps
        self._make()

    def _make(self) -> None:
        self.kernel._make()
        self._mean_fn = PosteriorMean(self.eps)
        self._var_fn = PosteriorVariance(self.eps, self.sigma_sq)
        self._fast_posterior_mean_fn = FastPosteriorMean()
        self._fast_precompute_fn = FastPrecomputeCoefficients(self.eps)

    def set_eps(self, **eps) -> None:
        """
        Reset :math:`\\varepsilon` value or bounds.

        Uses existing value and bounds as defaults.

        Args:
            eps:
                A hyperparameter dict.
        """
        self.eps._set(**eps)

    def fixed(self) -> bool:
        """
        Checks whether all kernel and model parameters are fixed.

        This is a convenience utility to determine whether optimization is
        required.

        Returns:
            Returns `True` if all parameters are fixed, and `False` otherwise.
        """
        for p in self.kernel._hyperparameters:
            if not self.kernel._hyperparameters[p].fixed():
                return False
        if not self.eps.fixed():
            return False
        return True

    def get_opt_params(
        self,
    ) -> Tuple[List[str], mm.ndarray, mm.ndarray]:
        """
        Return lists of unfixed hyperparameter names, values, and bounds.

        Returns
        -------
            names:
                A list of unfixed hyperparameter names.
            params:
                A list of unfixed hyperparameter values.
            bounds:
                A list of unfixed hyperparameter bound tuples.
        """
        names, params, bounds = self.kernel.get_opt_params()
        self.eps.append_lists("eps", names, params, bounds)
        return names, mm.array(params), mm.array(bounds)

    def fast_coefficients(
        self,
        K: mm.ndarray,
        train_nn_targets_fast: mm.ndarray,
    ) -> mm.ndarray:
        """
        Produces coefficient matrix for the fast posterior mean given in
        Equation (8) of [dunton2022fast]_.

        To form each row of this matrix, we compute

        .. math::
            \\mathbf{C}_{N^*}(i, :) =
                (K_{\\hat{\\theta}} (X_{N^*}, X_{N^*})
                + \\varepsilon I_k)^{-1} Y(X_{N^*}).

        Here :math:`X_{N^*}` is the union of the nearest neighbor of the ith
        test point and the `nn_count - 1` nearest neighbors of this nearest
        neighbor, :math:`K_{\\hat{\\theta}}` is the trained kernel functor
        specified by `self.kernel`, :math:`\\varepsilon I_k` is a diagonal
        noise matrix whose diagonal is the value of the
        `self.eps` hyperparameter, and :math:`Y(X_{N^*})` is the
        `(train_count,)` vector of responses corresponding to the
        training features indexed by $N^*$.

        Args:
            K:
                The full pairwise kernel tensor of shape
                `(train_count, nn_count, nn_count)`.
            train_nn_targets_fast:
                The nearest neighbor response of each
                training points of shape
                `(train_count, nn_count, response_count)`.
        Returns:
            A matrix of shape `(train_count, nn_count)` whose rows are
            the precomputed coefficients for fast posterior mean inference.

        """

        return self._fast_precompute_fn(
            K,
            train_nn_targets_fast,
        )

    def posterior_mean(
        self, K: mm.ndarray, Kcross: mm.ndarray, batch_nn_targets: mm.ndarray
    ) -> mm.ndarray:
        """
        Returns the posterior mean from the provided covariance,
        cross-covariance, and target tensors.

        Computes parallelized local solves of systems of linear equations using
        the last two dimensions of `K` along with `Kcross` and
        `batch_nn_targets` to predict responses in terms of the posterior mean.
        Assumes that kernel tensor `K` and cross-covariance
        matrix `Kcross` are already computed and given as arguments.

        Returns the predicted response in the form of a posterior
        mean for each element of the batch of observations, as computed in
        Equation (3.4) of [muyskens2021muygps]_. For each batch element
        :math:`\\mathbf{x}_i`, we compute

        .. math::
            \\widehat{Y}_{NN} (\\mathbf{x}_i \\mid X_{N_i}) =
                K_\\theta (\\mathbf{x}_i, X_{N_i})
                (K_\\theta (X_{N_i}, X_{N_i}) + \\varepsilon I_k)^{-1}
                Y(X_{N_i}).

        Here :math:`X_{N_i}` is the set of nearest neighbors of
        :math:`\\mathbf{x}_i` in the training data, :math:`K_\\theta` is the
        kernel functor specified by `self.kernel`, :math:`\\varepsilon I_k` is a
        diagonal homoscedastic noise matrix whose diagonal is the value of the
        `self.eps` hyperparameter, and :math:`Y(X_{N_i})` is the
        `(nn_count, response_count)` matrix of responses of the nearest
        neighbors given by the second two dimensions of the `batch_nn_targets`
        argument.

        Args:
            K:
                A tensor of shape `(batch_count, nn_count, nn_count)` containing
                the `(nn_count, nn_count` -shaped kernel matrices corresponding
                to each of the batch elements.
            Kcross:
                A matrix of shape `(batch_count, nn_count)` containing the
                `1 x nn_count` -shaped cross-covariance matrix corresponding
                to each of the batch elements.
            batch_nn_targets:
                A tensor of shape `(batch_count, nn_count, response_count)`
                whose last dimension lists the vector-valued responses for the
                nearest neighbors of each batch element.

        Returns:
            A matrix of shape `(batch_count, response_count)` whose rows are
            the predicted response for each of the given indices.
        """
        return self._mean_fn(K, Kcross, batch_nn_targets)

    def posterior_variance(
        self,
        K: mm.ndarray,
        Kcross: mm.ndarray,
    ) -> mm.ndarray:
        """
        Returns the posterior mean from the provided covariance and
        cross-covariance tensors.

        Return the local posterior variances of each prediction, corresponding
        to the diagonal elements of a covariance matrix. For each batch element
        :math:`\\mathbf{x}_i`, we compute

        .. math::
            Var(\\widehat{Y}_{NN} (\\mathbf{x}_i \\mid X_{N_i})) =
                K_\\theta (\\mathbf{x}_i, \\mathbf{x}_i) -
                K_\\theta (\\mathbf{x}_i, X_{N_i})
                (K_\\theta (X_{N_i}, X_{N_i}) + \\varepsilon I_k)^{-1}
                K_\\theta (X_{N_i}, \\mathbf{x}_i).

        Args:
            K:
                A tensor of shape `(batch_count, nn_count, nn_count)` containing
                the `(nn_count, nn_count` -shaped kernel matrices corresponding
                to each of the batch elements.
            Kcross:
                A matrix of shape `(batch_count, nn_count)` containing the
                `1 x nn_count` -shaped cross-covariance matrix corresponding
                to each of the batch elements.

        Returns:
            A vector of shape `(batch_count, response_count)` consisting of the
            diagonal elements of the posterior variance.
        """
        return self._var_fn(K, Kcross)

    def fast_posterior_mean(
        self,
        Kcross: mm.ndarray,
        coeffs_tensor: mm.ndarray,
    ) -> mm.ndarray:
        """
        Performs fast posterior mean inference using provided cross-covariance
        and precomputed coefficient matrix.

        Assumes that cross-covariance matrix `Kcross` is already computed and
        given as an argument.

        Returns the predicted response in the form of a posterior
        mean for each element of the batch of observations, as computed in
        Equation (9) of [dunton2022fast]_. For each test point
        :math:`\\mathbf{z}`, we compute

        .. math::
            \\widehat{Y} (\\mathbf{z} \\mid X) =
                K_\\theta (\\mathbf{z}, X_{N^*}) \\mathbf{C}_{N^*}.

        Here :math:`X_{N^*}` is the union of the nearest neighbor of the queried
        test point :math:`\\mathbf{z}` and the nearest neighbors of that
        training point, :math:`K_\\theta` is the kernel functor
        specified by `self.kernel`, and :math:`\\mathbf{C}_{N^*}` is
        the matrix of precomputed coefficients given in Equation (8)
        of [dunton2022fast]_.

        Args:
            Kcross:
                A matrix of shape `(batch_count, nn_count)` containing the
                `1 x nn_count` -shaped cross-covariance vector corresponding
                to each of the batch elements.
            coeffs_tensor:
                A matrix of shape `(batch_count, nn_count, response_count)`
                whose rows are given by precomputed coefficients.


        Returns:
            A matrix of shape `(batch_count, response_count)` whose rows are
            the predicted response for each of the given indices.
        """
        return self._fast_posterior_mean_fn(Kcross, coeffs_tensor)

    def apply_new_noise(
        self,
        new_noise: Union[HeteroscedasticNoise, HomoscedasticNoise, NullNoise],
    ):
        """
        Updates the homo/heteroscedastic noise parameter(s) of a MuyGPs model.
        To be used when the MuyGPs model has been trained and needs to be
        used for prediction, or if multiple batches are needed during training
        of a heteroscedastic model.

        Args:
            new_noise:
                If homoscedastic, a float to update the nugget parameter.
                If heteroscedastic, a matrix of shape
                `(test_count, nn_count)` containing the measurement
                noise corresponding to the nearest neighbors of each test point.
        Returns:
            A MuyGPs model with updated noise parameter(s).
        """
        ret = deepcopy(self)
        ret.eps = new_noise
        ret._mean_fn = PosteriorMean(ret.eps)
        ret._var_fn = PosteriorVariance(ret.eps, ret.sigma_sq)
        return ret

    def get_opt_mean_fn(self) -> Callable:
        """
        Return a posterior mean function for use in optimization.

        This function is designed for use with
        :func:`MuyGPyS.optimize.chassis.optimize_from_tensors()` and assumes
        that either `eps` will be passed via a keyword argument or not at all.

        Returns:
            A function implementing the posterior mean, where `eps` is either
            fixed or takes updating values during optimization. The function
            expects keyword arguments corresponding to current hyperparameter
            values for unfixed parameters.
        """
        return self._mean_fn.get_opt_fn()

    def get_opt_var_fn(self) -> Callable:
        """
        Return a posterior variance function for use in optimization.

        This function is designed for use with
        :func:`MuyGPyS.optimize.chassis.optimize_from_tensors()` and assumes
        that either `eps` will be passed via a keyword argument or not at all.

        Returns:
            A function implementing posterior variance, where `eps` is either
            fixed or takes updating values during optimization. The function
            expects keyword arguments corresponding to current hyperparameter
            values for unfixed parameters.
        """
        return self._var_fn.get_opt_fn()

    def __eq__(self, rhs) -> bool:
        if isinstance(rhs, self.__class__):
            return all(
                (
                    all(
                        self.kernel._hyperparameters[h]()
                        == rhs.kernel._hyperparameters[h]()
                        for h in self.kernel._hyperparameters
                    ),
                    self.eps() == rhs.eps(),
                    self.sigma_sq() == rhs.sigma_sq(),
                )
            )
        else:
            return False



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from jax import jit
from jax.scipy.special import gammaln
from tensorflow_probability.substrates import jax as tfp

import MuyGPyS._src.math.jax as jnp


@jit
def _rbf_fn(squared_dists: jnp.ndarray, **kwargs) -> jnp.ndarray:
    return jnp.exp(-squared_dists / 2)


@jit
def _matern_05_fn(dists: jnp.ndarray, **kwargs) -> jnp.ndarray:
    return jnp.exp(-dists)


@jit
def _matern_15_fn(dists: jnp.ndarray, **kwargs) -> jnp.ndarray:
    K = dists * jnp.sqrt(3)
    return (1.0 + K) * jnp.exp(-K)


@jit
def _matern_25_fn(dists: jnp.ndarray, **kwargs) -> jnp.ndarray:
    K = dists * jnp.sqrt(5)
    return (1.0 + K + K**2 / 3.0) * jnp.exp(-K)


@jit
def _matern_inf_fn(dists: jnp.ndarray, **kwargs) -> jnp.ndarray:
    return jnp.exp(-(dists**2) / 2.0)


@jit
def _matern_gen_fn(dists: jnp.ndarray, nu: float, **kwargs) -> jnp.ndarray:
    K = dists
    diag_indices = jnp.arange(K.shape[1])
    if len(K.shape) == 3:
        K = K.at[:, diag_indices, diag_indices].set(1.0)
    tmp = jnp.sqrt(2 * nu) * K
    const_val = (2 ** (1.0 - nu)) / jnp.exp(gammaln(nu))
    if len(K.shape) == 2:
        K = K.at[:, :].set(const_val)
    elif len(K.shape) == 3:
        K = K.at[:, :, :].set(const_val)
    K *= tmp**nu
    K *= tfp.math.bessel_kve(nu, tmp) / jnp.exp(jnp.abs(tmp))
    if len(K.shape) == 3:
        K = K.at[:, jnp.arange(K.shape[1]), jnp.arange(K.shape[1])].set(1.0)
    return K



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

import warnings
from typing import Callable

import MuyGPyS._src.math.numpy as np
from MuyGPyS import config

world = config.mpi_state.comm_world
if world is not None:
    rank = config.mpi_state.comm_world.Get_rank()
    size = config.mpi_state.comm_world.Get_size()
else:
    rank = 0
    size = 1


def _rank0() -> bool:
    return True if rank == 0 else False


def _print0(*args, **kwargs):
    if _rank0() is True:
        print(*args)


def _warn0(*args, **kwargs):
    if _rank0() is True:
        warnings.warn(*args, **kwargs)


def _get_chunk_sizes(count: int, size: int):
    floor = int(count / size)
    remainder = count - floor * size
    return [
        floor + 1 if i >= (size - remainder) else floor for i in range(size)
    ]


def _prepare_parallel_data(size, chunk_sizes, *args):
    count = len(args)
    ret = [list() for _ in range(count)]
    offsets = np.array(
        [np.sum(chunk_sizes[:i]) for i in range(size)], dtype=int
    )
    for i in range(size):
        for j, arg in enumerate(args):
            ret[j].append(arg[offsets[i] : offsets[i] + chunk_sizes[i]])
    return ret


def _big_scatter(chunks, root=0):
    """
    Based upon this reply from a mpi4py dev
    https://github.com/mpi4py/mpi4py/issues/119#issuecomment-945390731
    """
    if size == 0:
        my_chunk = chunks
    elif rank == 0:
        for i, chunk in enumerate(chunks):
            if i == 0:
                my_chunk = chunk
            else:
                world.send(chunk, dest=i)
    else:
        my_chunk = world.recv(source=0)
    return my_chunk


def _chunk_tensor(tensors, return_count=1):
    if rank == 0:
        if return_count == 1:
            tensors = [tensors]
        data_count = tensors[0].shape[0]
        chunk_sizes = _get_chunk_sizes(data_count, size)
        tensor_chunks = _prepare_parallel_data(size, chunk_sizes, *tensors)
    else:
        tensor_chunks = [None for _ in range(return_count)]
    local_chunks = list()
    for chunks in tensor_chunks:
        local_chunks.append(_big_scatter(chunks, root=0))
    if return_count == 1:
        local_chunks = local_chunks[0]
    return local_chunks


def _chunk_function_tensor(func: Callable, *args, return_count=1):
    if rank == 0:
        tensors = func(*args)
    else:
        tensors = None
    return _chunk_tensor(tensors, return_count=return_count)


def _consistent_unchunk_tensor(tensor: np.ndarray) -> np.ndarray:
    """
    If we are using an MPI implementation, allgather the tensor across all
    cores. Otherwise NOOP.

    The purpose of this function is to allow the existing serial testing harness
    to also test the mpi implementations without the need for additional codes.

    Args:
        tensor:
            A tensor, which might be a simple serial tensors or distributed
            chunks if it is the product of the mpi implementation.

    Return:
        The same tensor if a serial implementation, else an allgathered tensor
        of the distributed chunks.
    """
    if tensor is None:
        return tensor
    if _is_mpi_mode() is True:
        if len(tensor.shape) > 1:
            return np.vstack(config.mpi_state.comm_world.allgather(tensor))
        else:
            return np.concatenate(config.mpi_state.comm_world.allgather(tensor))
    else:
        return tensor


def _consistent_chunk_tensor(tensor) -> np.ndarray:
    if _is_mpi_mode() is True:
        return _chunk_tensor(tensor)
    else:
        return tensor


def _consistent_reduce_scalar(scalar):
    if _is_mpi_mode() is True:
        from mpi4py import MPI

        return world.allreduce(scalar, op=MPI.SUM)
    else:
        return scalar


def _is_mpi_mode() -> bool:
    return config.state.backend == "mpi"



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

"""
Hyperparameters

Hyperparameters specifications are expected to be provided in `Dict` form with
the keys `"val"` and `"bounds"`. `"bounds"` is either a 2-tuple indicating a
lower and upper optimization bound or the string `"fixed"`, which exempts the
hyperparameter from optimization (`"fixed"` is the default behavior if `bounds`
is unspecified). `"val"` is either a floating point value (within the range
between the upper and lower bounds if specified).
"""

from collections.abc import Sequence
from numbers import Number
from typing import Callable, cast, List, Tuple, Type, Union

import MuyGPyS._src.math.numpy as np
import MuyGPyS._src.math as mm
from MuyGPyS import config
from MuyGPyS._src.mpi_utils import _is_mpi_mode


class ScalarHyperparameter:
    """
    A MuyGPs kernel or model Hyperparameter.

    Hyperparameters are defined by a value and optimization bounds. Values must
    be scalar numeric types, and bounds are either a len == 2 iterable container
    whose elements are numeric scalars in increasing order, or the string
    `fixed`. If `bounds == "fixed"` (the default behavior), the hyperparameter
    value will remain fixed during optimization. `val` must remain within the
    range of the upper and lower bounds, if not `fixed`.

    Args:
        val:
            A scalar within the range of the upper and lower bounds (if given).
            val can also be the strings `"sample"` or `"log_sample"`, which will
            result in randomly sampling a value within the range given by the
            bounds.
        bounds:
            Iterable container of len 2 containing lower and upper bounds (in
            that order), or the string `"fixed"`.

    Raises:
        ValueError:
            Any `bounds` string other than `"fixed"` will produce an error.
        ValueError:
            A non-iterable non-string type for `bounds` will produce an
            error.
        ValueError:
            A `bounds` iterable of len other than 2 will produce an error.
        ValueError:
            Iterable `bounds` values of non-numeric types will produce an error.
        ValueError:
            A lower bound that is not less than an upper bound will produce
            an error.
        ValueError:
            `val == "sample" or val == "log_sample"` will produce an error
            if `self._bounds == "fixed"`.
        ValueError:
            Any string other than `"sample"` or `"log_sample"` will produce
            an error.
        ValueError:
            A `val` outside of the range specified by `self._bounds` will
            produce an error.
    """

    def __init__(
        self,
        val: Union[str, float],
        bounds: Union[str, Tuple[float, float]] = "fixed",
    ):
        """
        Initialize a hyperparameter.
        """
        self._set_bounds(bounds)
        self._set_val(val)

    def __str__(self, **kwargs):
        bstring = "fixed" if self._fixed is True else self._bounds
        return f"{type(self).__name__}({self._val}, {bstring})"

    def _set(self, rhs) -> None:
        """
        Reset hyperparameter value and/or bounds using keyword arguments.

        Args:
            rhs:
                Another hyperparameter.
        Raises:
            ValueError:
                Any `bounds` string other than `"fixed"` will produce an error.
            ValueError:
                A non-numeric, non-numeric, or non-string type for `bounds` will
                produce an error.
            ValueError:
                A `bounds` iterable of len other than 2 will produce an error.
            ValueError:
                Iterable `bounds` values of non-numeric types will produce an
                error.
            ValueError:
                A lower bound that is not less than an upper bound will produce
                an error.
            ValueError:
                `val == "sample" or val == "log_sample"` will produce an error
                if `self._bounds == "fixed"`.
            ValueError:
                Any string other than `"sample"` or `"log_sample"` will produce
                an error.
            ValueError:
                A `val` outside of the range specified by `self._bounds` will
                produce an error.
        """
        self._val = rhs._val
        self._bounds = rhs._bounds
        self._fixed = rhs._fixed

    def _sample_val(self, val: str) -> float:
        if self.fixed() is True:
            if isinstance(val, str):
                raise ValueError(
                    f"Fixed bounds do not support string value ({val}) prompts."
                )
        if val == "sample":
            newval = float(
                np.random.uniform(low=self._bounds[0], high=self._bounds[1])
            )
        elif val == "log_sample":
            newval = float(
                np.exp(
                    np.random.uniform(
                        low=np.log(self._bounds[0]),
                        high=np.log(self._bounds[1]),
                    )
                )
            )
        else:
            raise ValueError(f"Unsupported string hyperparameter value {val}.")
        if _is_mpi_mode() is True:
            newval = config.mpi_state.comm_world.bcast(newval, root=0)
        return newval

    def _set_val(self, val: Union[str, float]) -> None:
        """
        Set hyperparameter value; sample if appropriate.

        Throws on out-of-range and other badness.

        Args:
            val:
                A valid scalar value or the strings `"sample"` or
                `"log_sample"`.

        Raises:
            ValueError:
                A non-scalar, non-numeric and non-string val will produce an
                error.
            ValueError:
                `val == "sample" or val == "log_sample"` will produce an error
                if `bounds == "fixed"`.
            ValueError:
                Any `val` string other than `"sample"` or `"log_sample"` will
                produce an error.
            ValueError:
                A `val` outside of the range specified by `bounds` will
                produce an error.
        """
        if isinstance(val, str):
            val = self._sample_val(val)
        if isinstance(val, Sequence) or hasattr(val, "__len__"):
            raise ValueError(
                f"Nonscalar hyperparameter value {val} is not allowed."
            )
        if not isinstance(val, mm.ndarray):
            val = float(val)

        if self.fixed() is False:
            any_below = np.any(
                np.choose(
                    cast(float, val) < cast(float, self._bounds[0]) - 1e-5,
                    [False, True],
                )
            )
            any_above = np.any(
                np.choose(
                    cast(float, val) > cast(float, self._bounds[1]) + 1e-5,
                    [False, True],
                )
            )
            if any_below:
                raise ValueError(
                    f"Hyperparameter value {val} is lesser than the "
                    f"optimization lower bound {self._bounds[0]}"
                )
            if any_above:
                raise ValueError(
                    f"Hyperparameter value {val} is greater than the "
                    f"optimization upper bound {self._bounds[1]}"
                )
        self._val = mm.parameter(val)

    def _set_bounds(
        self,
        bounds: Union[str, Tuple[float, float]],
    ) -> None:
        """
        Set hyperparameter bounds.

        Args:
            bounds:
                Iterable container of len 2 containing lower and upper bounds
                (in that order), or the string `"fixed"`.

        Raises:
            ValueError:
                Any string other than `"fixed"` will produce an error.
            ValueError:
                A non-iterable type will produce an error.
            ValueError:
                An iterable of len other than 2 will produce an error.
            ValueError:
                Iterable values of non-numeric types will produce an error.
            ValueError:
                A lower bound that is not less than an upper bound will produce
                an error.
        """
        if isinstance(bounds, str) is True:
            if bounds == "fixed":
                self._bounds = (0.0, 0.0)  # default value
                self._fixed = True
            else:
                raise ValueError(f"Unknown bound option {bounds}.")
        else:
            if hasattr(bounds, "__iter__") is not True:
                raise ValueError(
                    f"Unknown bound optiom {bounds} of a non-iterable type "
                    f"{type(bounds)}."
                )
            if len(bounds) != 2:
                raise ValueError(
                    f"Provided hyperparameter optimization bounds have "
                    f"unsupported length {len(bounds)}."
                )
            if isinstance(bounds[0], Number) is not True:
                raise ValueError(
                    f"Nonscalar {bounds[0]} of type {type(bounds[0])} is not a "
                    f"supported hyperparameter bound type."
                )
            if isinstance(bounds[1], Number) is not True:
                raise ValueError(
                    f"Nonscalar {bounds[1]} of type {type(bounds[1])} is not a "
                    f"supported hyperparameter bound type."
                )
            bounds = (float(bounds[0]), float(bounds[1]))
            if bounds[0] > bounds[1]:
                raise ValueError(
                    f"Lower bound {bounds[0]} is not lesser than upper bound "
                    f"{bounds[1]}."
                )
            self._bounds = bounds
            self._fixed = False

    def __call__(self) -> float:
        """
        Value accessor.

        Returns:
            The current value of the hyperparameter.
        """
        return self._val

    def get_bounds(self) -> Tuple[float, float]:
        """
        Bounds accessor.

        Returns:
            The lower and upper bound tuple.
        """
        return self._bounds

    def fixed(self) -> bool:
        """
        Report whether the parameter is fixed, and is to be ignored during
        optimization.

        Returns:
            `True` if fixed, `False` otherwise.
        """
        return self._fixed

    def apply(self, fn: Callable, name: str) -> Callable:
        if self.fixed():

            def applied_fn(*args, **kwargs):
                kwargs.setdefault(name, self())
                return fn(*args, **kwargs)

            return applied_fn

        return fn

    def append_lists(
        self,
        name: str,
        names: List[str],
        params: List[float],
        bounds: List[Tuple[float, float]],
    ):
        if not self.fixed():
            names.append(name)
            params.append(self())
            bounds.append(self.get_bounds())


def _init_scalar_hyperparameter(
    val_def: Union[str, float],
    bounds_def: Union[str, Tuple[float, float]],
    type: Type = ScalarHyperparameter,
    **kwargs,
) -> ScalarHyperparameter:
    """
    Initialize a hyperparameter given default values.

    Args:
        val:
            A valid value or `"sample"` or `"log_sample"`.
        bounds:
            Iterable container of len 2 containing lower and upper bounds (in
            that order), or the string `"fixed"`.
        kwargs:
            A hyperparameter dict including as subset of the keys `val` and
            `bounds`.
    """
    val = kwargs.get("val", val_def)
    bounds = kwargs.get("bounds", bounds_def)
    return type(val, bounds)



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

"""
Resources and high-level API for a fast posterior mean inference workflow.

:func:`~MuyGPyS.examples.fast_posterior_mean.make_fast_regressor` is a
high-level API for creating the necessary components for fast posterior mean
inference.
:func:`~MuyGPyS.examples.fast_posterior_mean.make_fast_multivariate_regressor`
is a high-level API for creating the necessary components for fast posterior
mean inference with multiple outputs.

:func:`~MuyGPyS.examples.fast_posterior_mean.do_fast_posterior_mean` is a
high-level api for executing a simple, generic fast posterior medan workflow
given data.
It calls the maker APIs above and
:func:`~MuyGPyS.examples.fast_posterior_mean.fast_posterior_mean_any`.
"""

from time import perf_counter
from typing import Callable, Dict, List, Optional, Tuple, Union

import MuyGPyS._src.math as mm
from MuyGPyS.gp import MuyGPS, MultivariateMuyGPS as MMuyGPS
from MuyGPyS.examples.from_indices import fast_posterior_mean_from_indices
from MuyGPyS.examples.regress import _decide_and_make_regressor
from MuyGPyS.gp.tensors import fast_nn_update
from MuyGPyS.gp.tensors import pairwise_tensor
from MuyGPyS.neighbors import NN_Wrapper
from MuyGPyS.optimize.loss import lool_fn


def make_fast_regressor(
    muygps: MuyGPS,
    nbrs_lookup: NN_Wrapper,
    train_features: mm.ndarray,
    train_targets: mm.ndarray,
) -> Tuple[mm.ndarray, mm.ndarray]:
    """
    Convenience function for creating precomputed coefficient matrix and neighbor lookup data
    structure.

    Args:
        muygps:
            A (possibly trained) MuyGPS object.
        nbrs_lookup:
             A data structure supporting nearest neighbor queries into
            `train_features`.
        train_features:
            A matrix of shape `(train_count, feature_count)` whose rows consist
            of observation vectors of the train data.
        train_targets:
            A matrix of shape `(train_count, response_count)` whose rows consist
            of response vectors of the train data.

    Returns
    -------
    precomputed_coefficients_matrix:
        A matrix of shape `(train_count, nn_count)` whose rows list the
        precomputed coefficients for each nearest neighbors set in the
        training data.
    nn_indices:
        A numpy.ndarrray supporting nearest neighbor queries.
    """

    num_training_samples, _ = train_features.shape
    nn_indices, _ = nbrs_lookup.get_batch_nns(
        mm.arange(0, num_training_samples)
    )
    nn_indices = fast_nn_update(nn_indices)

    train_nn_targets = train_targets[nn_indices]
    K = muygps.kernel(pairwise_tensor(train_features, nn_indices))

    precomputed_coefficients_matrix = muygps.fast_coefficients(
        K, train_nn_targets
    )

    return precomputed_coefficients_matrix, nn_indices


def make_fast_multivariate_regressor(
    mmuygps: MMuyGPS,
    nbrs_lookup: NN_Wrapper,
    train_features: mm.ndarray,
    train_targets: mm.ndarray,
) -> Tuple[mm.ndarray, mm.ndarray]:
    """
    Convenience function for creating precomputed coefficient matrix and neighbor lookup data
    structure.

    Args:
        muygps:
            A trained MultivariateMuyGPS object.
        nbrs_lookup:
             A data structure supporting nearest neighbor queries into
            `train_features`.
        train_features:
            A matrix of shape `(train_count, feature_count)` whose rows consist
            of observation vectors of the train data.
        train_targets:
            A matrix of shape `(train_count, response_count)` whose rows consist
            of response vectors of the train data.

    Returns
    -------
    precomputed_coefficients_matrix:
        A matrix of shape `(train_count, nn_count)` whose rows list the
        precomputed coefficients for each nearest neighbors set in the
        training data.
    nn_indices:
        An array supporting nearest neighbor queries.
    """
    num_training_samples, _ = train_features.shape
    nn_indices, _ = nbrs_lookup.get_batch_nns(
        mm.arange(0, num_training_samples)
    )

    nn_indices = fast_nn_update(nn_indices)
    pairwise_diffs_fast = pairwise_tensor(train_features, nn_indices)
    train_nn_targets = train_targets[nn_indices]
    precomputed_coefficients_matrix = mmuygps.fast_coefficients(
        pairwise_diffs_fast, train_nn_targets
    )
    return precomputed_coefficients_matrix, nn_indices


def _decide_and_make_fast_regressor(
    muygps: Union[MuyGPS, MMuyGPS],
    nbrs_lookup: NN_Wrapper,
    train_features: mm.ndarray,
    train_targets: mm.ndarray,
) -> Tuple[mm.ndarray, mm.ndarray]:
    if isinstance(muygps, MuyGPS):
        precomputed_coefficients_matrix, nn_indices = make_fast_regressor(
            muygps,
            nbrs_lookup,
            train_features,
            train_targets,
        )
    else:
        (
            precomputed_coefficients_matrix,
            nn_indices,
        ) = make_fast_multivariate_regressor(
            muygps,
            nbrs_lookup,
            train_features,
            train_targets,
        )
    return precomputed_coefficients_matrix, nn_indices


def do_fast_posterior_mean(
    test_features: mm.ndarray,
    train_features: mm.ndarray,
    train_targets: mm.ndarray,
    nn_count: int = 30,
    batch_count: int = 200,
    loss_fn: Callable = lool_fn,
    obj_method: str = "loo_crossval",
    opt_method: str = "bayes",
    sigma_method: Optional[str] = "analytic",
    k_kwargs: Union[Dict, Union[List[Dict], Tuple[Dict, ...]]] = dict(),
    nn_kwargs: Dict = dict(),
    opt_kwargs: Dict = dict(),
    verbose: bool = False,
) -> Tuple[mm.ndarray, NN_Wrapper, mm.ndarray, mm.ndarray, Dict]:
    """
    Convenience function initializing a model and performing fast posterior mean
    inference.

    Expected parameters include keyword argument dicts specifying kernel
    parameters and nearest neighbor parameters. See the docstrings of the
    appropriate functions for specifics.

    Also supports workflows relying upon multivariate models. In order to create
    a multivariate model, specify the `kern` argument and pass a list of
    hyperparameter dicts to `k_kwargs`.

    Example:
        >>> from MuyGPyS.testing.test_utils import _make_gaussian_data
        >>> from MuyGPyS.examples.fast_posterior_mean import do_fast_posterior_mean
        >>> from MuyGPyS.optimize.objective import mse_fn
        >>> train, test = _make_gaussian_data(10000, 1000, 100, 10)
        >>> nn_kwargs = {"nn_method": "exact", "algorithm": "ball_tree"}
        >>> k_kwargs = {
        ...         "kern": "rbf",
        ...         "metric": "F2",
        ...         "eps": {"val": 1e-5},
        ...         "length_scale": {"val": 1.0, "bounds": (1e-2, 1e2)}
        ... }
        >>> muygps, nbrs_lookup, predictions, precomputed_coefficients_matrix
        ...         = do_fast_posterior_mean(
        ...         test['input'],
        ...         train['input'],
        ...         train['output'],
        ...         nn_count=30,
        ...         batch_count=200,
        ...         loss_fn=lool_fn,
        ...         obj_method="loo_crossval",
        ...         opt_method="bayes",
        ...         k_kwargs=k_kwargs,
        ...         nn_kwargs=nn_kwargs,
        ...         verbose=False,
        ... )
    Args:
        test_features:
            A matrix of shape `(test_count, feature_count)` whose rows consist
            of observation vectors of the test data.
        train_features:
            A matrix of shape `(train_count, feature_count)` whose rows consist
            of observation vectors of the train data.
        train_targets:
            A matrix of shape `(train_count, response_count)` whose rows consist
            of response vectors of the train data.
        nn_count:
            The number of nearest neighbors to employ.
        batch_count:
            The number of elements to sample batch for hyperparameter
            optimization.
        loss_fn:
            The loss functor to use in hyperparameter optimization. Ignored if
            all of the parameters specified by argument `k_kwargs` are fixed.
        obj_method:
            Indicates the objective function to be minimized. Currently
            restricted to `"loo_crossval"`.
        opt_method:
            Indicates the optimization method to be used. Currently restricted
            to `"bayesian"` and `"scipy"`.
        sigma_method:
            The optimization method to be employed to learn the `sigma_sq`
            hyperparameter. Currently supports only `"analytic"` and `None`. If
            the value is not `None`, the returned
            :class:`MuyGPyS.gp.muygps.MuyGPS` object will possess a `sigma_sq`
            member whose value, invoked via `muygps.sigma_sq()`, is a
            `(response_count,)` vector to be used for scaling posterior
            variances.
        k_kwargs:
            If given a list or tuple of length `response_count`, assume that the
            elements are dicts containing kernel initialization keyword
            arguments for the creation of a multivariate model (see
            :func:`~MuyGPyS.examples.regress.make_multivariate_regressor`).
            If given a dict, assume that the elements are keyword arguments to
            a MuyGPs model (see
            :func:`~MuyGPyS.examples.regress.make_regressor`).
        nn_kwargs:
            Parameters for the nearest neighbors wrapper. See
            :class:`MuyGPyS.neighbors.NN_Wrapper` for the supported methods and
            their parameters.
        opt_kwargs:
            Parameters for the wrapped optimizer. See the docs of the
            corresponding library for supported parameters.
        verbose:
            If `True`, print summary statistics.

    Returns
    -------
    muygps:
        A (possibly trained) MuyGPs object.
    nbrs_lookup:
        A data structure supporting nearest neighbor queries into
        `train_features`.
    predictions:
        The predicted response associated with each test observation.
    precomputed_coefficients_matrix:
        A matrix of shape `(train_count, nn_count)` whose rows list the
        precomputed coefficients for each nearest neighbors set in the
        training data.
    timing:
        A dictionary containing timings for the training, precomputation,
        nearest neighbor computation, and prediction.
    """
    regressor, nbrs_lookup = _decide_and_make_regressor(
        train_features,
        train_targets,
        nn_count=nn_count,
        batch_count=batch_count,
        loss_fn=loss_fn,
        obj_method=obj_method,
        opt_method=opt_method,
        sigma_method=sigma_method,
        k_kwargs=k_kwargs,
        nn_kwargs=nn_kwargs,
        opt_kwargs=opt_kwargs,
        verbose=verbose,
    )

    (
        posterior_mean,
        precomputed_coefficients_matrix,
        timing,
    ) = fast_posterior_mean_any(
        regressor,
        test_features,
        train_features,
        nbrs_lookup,
        train_targets,
    )
    return (
        regressor,
        nbrs_lookup,
        posterior_mean,
        precomputed_coefficients_matrix,
        timing,
    )


def fast_posterior_mean_any(
    muygps: Union[MuyGPS, MMuyGPS],
    test_features: mm.ndarray,
    train_features: mm.ndarray,
    nbrs_lookup: NN_Wrapper,
    train_targets: mm.ndarray,
) -> Tuple[mm.ndarray, mm.ndarray, Dict]:
    """
    Convenience function performing fast posterior mean inference using a
    pre-trained model.

    Also supports workflows relying upon multivariate models.

    Args:
        muygps:
            A (possibly trained) MuyGPS object.
        test_features:
            A matrix of shape `(test_count, feature_count)` whose rows consist
            of observation vectors of the test data.
        train_features:
            A matrix of shape `(train_count, feature_count)` whose rows consist
            of observation vectors of the train data.
        nbrs_lookup:
             A data structure supporting nearest neighbor queries into
            `train_features`.
        train_targets:
            A matrix of shape `(train_count, response_count)` whose rows consist
            of response vectors of the train data.


    Returns
    -------
    posterior_mean:
        The predicted response associated with each test observation.
    precomputed_coefficients_matrix:
        A matrix of shape `(train_count, nn_count)` whose rows list the
        precomputed coefficients for each nearest neighbors set in the
        training data.
    timing:
        A dictionary containing timings for the training, precomputation,
        nearest neighbor computation, and prediction.

    """
    time_start = perf_counter()
    (
        precomputed_coefficients_matrix,
        nn_indices,
    ) = _decide_and_make_fast_regressor(
        muygps,
        nbrs_lookup,
        train_features,
        train_targets,
    )
    time_precomp = perf_counter()

    time_agree = perf_counter()
    nn_indices = fast_nn_update(nn_indices)

    test_neighbors, _ = nbrs_lookup.get_nns(test_features)
    time_nn = perf_counter()

    closest_neighbor = test_neighbors[:, 0]
    closest_set_new = nn_indices[closest_neighbor, :].astype(int)
    num_test_samples, _ = test_features.shape

    posterior_mean = fast_posterior_mean_from_indices(
        muygps,
        mm.arange(0, num_test_samples),
        closest_set_new,
        test_features,
        train_features,
        closest_neighbor,
        precomputed_coefficients_matrix,
    )
    time_pred = perf_counter()

    timing = {
        "precompute": time_precomp - time_start,
        "agree": time_agree - time_precomp,
        "nn": time_nn - time_agree,
        "pred": time_pred - time_nn,
    }

    return posterior_mean, precomputed_coefficients_matrix, timing



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

"""
Loss Function Handling

MuyGPyS includes predefined loss functions and convenience functions for
indicating them to optimization.
"""

from typing import Callable

import MuyGPyS._src.math as mm
from MuyGPyS._src.optimize.loss import (
    _mse_fn,
    _cross_entropy_fn,
    _lool_fn,
    _lool_fn_unscaled,
    _pseudo_huber_fn,
    _looph_fn,
)
from MuyGPyS.optimize.utils import _switch_on_loss_method


def get_loss_func(loss_method: str) -> Callable:
    """
    Select a loss function based upon string key.

    Currently supports strings `"log"` or `"cross-entropy"` for
    :func:`MuyGPyS.optimize.objective.cross_entropy_fn` and `"mse"` for
    :func:`MuyGPyS.optimize.objective.mse_fn`.

    Args:
        predictions:
            The predicted response of shape `(batch_count, response_count)`.
        targets:
            The expected response of shape `(batch_count, response_count)`.

    Returns:
        The loss function Callable.

    Raises:
        NotImplementedError:
            Unrecognized strings will result in an error.
    """
    return _switch_on_loss_method(
        loss_method,
        lambda: cross_entropy_fn,
        lambda: mse_fn,
        lambda: lool_fn,
        lambda: pseudo_huber_fn,
        lambda: looph_fn,
    )


def cross_entropy_fn(
    predictions: mm.ndarray,
    targets: mm.ndarray,
) -> float:
    """
    Cross entropy function.

    Computes the cross entropy loss the predicted versus known response.
    Transforms `predictions` to be row-stochastic, and ensures that `targets`
    contains no negative elements.

    @NOTE[bwp] I don't remember why we hard-coded eps=1e-6. Might need to
    revisit.

    Args:
        predictions:
            The predicted response of shape `(batch_count, response_count)`.
        targets:
            The expected response of shape `(batch_count, response_count)`.

    Returns:
        The cross-entropy loss of the prediction.
    """
    return _cross_entropy_fn(predictions, targets, ll_eps=1e-6)


def mse_fn(
    predictions: mm.ndarray,
    targets: mm.ndarray,
) -> float:
    """
    Mean squared error function.

    Computes mean squared error loss of the predicted versus known response.
    Treats multivariate outputs as interchangeable in terms of loss penalty. The
    function computes

    .. math::
        \\ell_\\textrm{MSE}(f(x), y) = \\frac{1}{b} \\sum_{i=1}^b (f(x_i) - y)^2

    Args:
        predictions:
            The predicted response of shape `(batch_count, response_count)`.
        targets:
            The expected response of shape `(batch_count, response_count)`.

    Returns:
        The mse loss of the prediction.
    """
    return _mse_fn(predictions, targets)


def lool_fn(
    predictions: mm.ndarray,
    targets: mm.ndarray,
    variances: mm.ndarray,
    sigma_sq: mm.ndarray,
) -> float:
    """
    Leave-one-out likelihood function.

    Computes leave-one-out likelihood (LOOL) loss of the predicted versus known
    response. Treats multivariate outputs as interchangeable in terms of loss
    penalty. The function computes

    .. math::
        \\ell_\\textrm{lool}(f(x), y \\mid \\sigma^2) =
        \\sum_{i=1}^b \\sum_{j=1}^s
        \\left ( \\frac{(f(x_i) - y)}{\\sigma_j} \\right )^2 + \\log \\sigma_j^2

    Args:
        predictions:
            The predicted response of shape `(batch_count, response_count)`.
        targets:
            The expected response of shape `(batch_count, response_count)`.
        variances:
            The unscaled variance of the predicted responses of shape
            `(batch_count, response_count)`.
        sigma_sq:
            The sigma_sq variance scaling parameter of shape
            `(response_count,)`.

    Returns:
        The LOOL loss of the prediction.
    """
    return _lool_fn(predictions, targets, variances, sigma_sq)


def lool_fn_unscaled(
    predictions: mm.ndarray, targets: mm.ndarray, variances: mm.ndarray
) -> float:
    """
    Leave-one-out likelihood function.

    Computes leave-one-out likelihood (LOOL) loss of the predicted versus known
    response. Treats multivariate outputs as interchangeable in terms of loss
    penalty. Unlike lool_fn, does not require sigma_sq as an argument. The
    function computes

    .. math::
        \\ell_\\textrm{lool}(f(x), y \\mid \\sigma^2) = \\sum_{i=1}^b
        \\left ( \\frac{(f(x_i) - y)}{\\sigma_i} \\right )^2 + \\log \\sigma_i^2

    Args:
        predictions:
            The predicted response of shape `(batch_count, response_count)`.
        targets:
            The expected response of shape `(batch_count, response_count)`.
        variances:
            The unscaled variance of the predicted responses of shape
            `(batch_count, response_count)`.

    Returns:
        The LOOL loss of the prediction.
    """
    return _lool_fn_unscaled(predictions, targets, variances)


def pseudo_huber_fn(
    predictions: mm.ndarray, targets: mm.ndarray, boundary_scale: float = 1.5
) -> float:
    """
    Pseudo-Huber loss function.

    Computes a smooth approximation to the Huber loss function, which balances
    sensitive squared-error loss for relatively small errors and
    robust-to-outliers absolute loss for larger errors, so that the loss is not
    overly sensitive to outliers. Used the form from
    [wikipedia](https://en.wikipedia.org/wiki/Huber_loss#Pseudo-Huber_loss_function).
    The function computes

    .. math::
        \\ell_\\textrm{Pseudo-Huber}(f(x), y \\mid \\delta) =
            \\sum_{i=1}^b \\delta^2 \\left ( \\sqrt{
            1 + \\left ( \\frac{y_i - f(x_i)}{\\delta} \\right )^2
            } - 1 \\right )

    Args:
        predictions:
            The predicted response of shape `(batch_count, response_count)`.
        targets:
            The expected response of shape `(batch_count, response_count)`.
        boundary_scale:
            The boundary value for the residual beyond which the loss becomes
            approximately linear. Useful values depend on the scale of the
            response.

    Returns:
        The sum of pseudo-Huber losses of the predictions.
    """
    return _pseudo_huber_fn(predictions, targets, boundary_scale=boundary_scale)


def looph_fn(
    predictions: mm.ndarray,
    targets: mm.ndarray,
    variances: mm.ndarray,
    sigma_sq: mm.ndarray,
    boundary_scale: float = 1.5,
) -> float:
    """
    Variance-regularized pseudo-Huber loss function.

    Computes a smooth approximation to the Huber loss function, similar to
    :func:`pseudo_huber_fn`, with the addition of both a variance scaling and a
    additive logarithmic variance regularization term to avoid exploding the
    variance. The function computes

    .. math::
        \\ell_\\textrm{lool}(f(x), y \\mid \\delta, \\sigma^2) =
            \\sum_{i=1}^b \\delta^2 \\left ( \\sqrt{
            1 + \\left ( \\frac{y_i - f(x_i)}{\\delta \\sigma_i^2} \\right )^2
            } - 1 \\right ) + \\log \\sigma_i^2

    Args:
        predictions:
            The predicted response of shape `(batch_count, response_count)`.
        targets:
            The expected response of shape `(batch_count, response_count)`.
        variances:
            The unscaled variance of the predicted responses of shape
            `(batch_count, response_count)`.
        sigma_sq:
            The sigma_sq variance scaling parameter of shape
            `(response_count,)`.
        boundary_scale:
            The boundary value for the residual beyond which the loss becomes
            approximately linear. Useful values depend on the scale of the
            response.

    Returns:
        The sum of pseudo-Huber losses of the predictions.
    """
    return _looph_fn(
        predictions, targets, variances, sigma_sq, boundary_scale=boundary_scale
    )
